# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import datetime
from reda.importers.eit_version_2010 import _average_swapped_current_injections


def _extract_adc_data(mat, **kwargs):
    """Extract adc-channel related data (i.e., data that is captured for all 48
    channels of the 40-channel medusa system

    """
    md = mat['MD'].squeeze()
    frequencies = mat['MP']['fm'].take(0)

    # it seems that there exist different file formats under this same official
    # version.
    if md['fm'].size == frequencies.size:
        use_v = 0
    else:
        use_v = 1

    # print('@@@')
    # import IPython
    # IPython.embed()
    # exit()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, frequencies.size):
        frequency = frequencies[f_id]

        if use_v == 0:
            def get_field(key):
                return md[key][f_id]
        elif use_v == 1:
            def get_field(key):
                indices = np.where(
                    md['fm'].take(0) == frequencies[f_id])
                return md[key].take(0)[indices]

        # def get_field(key):
        #     indices = np.where(md['fm'].take(f_id) == frequencies[f_id])
        #     return md[key].take(f_id)[indices]

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in get_field('Time')]
        ).T.squeeze()

        column_names = ['ch{:02}'.format(i) for i in range(48)]
        ab = get_field('cni')

        index_pairs = [
            (channel, 'Ug3_{}'.format(i)) for channel in column_names
            for i in range(3)
        ]

        Ug3 = get_field('Ug3')
        ug3_reshaped = Ug3.reshape([Ug3.shape[0], Ug3.shape[1] * 3])
        df = pd.DataFrame(
            ug3_reshaped,
            index=pd.MultiIndex.from_arrays(
                [
                    ab[:, 0],
                    ab[:, 1],
                    np.ones(ab.shape[0]) * frequency,
                    timestamp
                ],
                names=['a', 'b', 'frequency', 'datetime']
            ),
            columns=pd.MultiIndex.from_tuples(
               index_pairs, names=['channel', 'parameter'])
        )
        dfl.append(df)

    dfl = pd.concat(dfl)
    dfl.sort_index(axis=0, inplace=True)
    dfl.sort_index(axis=1, inplace=True)

    return dfl


def _extract_md(mat, **kwargs):
    """Note that the md struct for this version is structured differently than
    the others...
    """
    md = mat['MD'].squeeze()
    frequencies = mat['MP']['fm'].take(0)

    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, frequencies.size):

        def get_field(key):
            indices = np.where(
                md['fm'].take(0) == frequencies[f_id])
            return md[key].take(0)[indices]

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in get_field('Time')]
        ).T.squeeze()

        df = pd.DataFrame()
        df['datetime'] = timestamp
        ab = get_field('cni')
        df['a'] = ab[:, 0]
        df['b'] = ab[:, 1]
        df['U0'] = get_field('U0')
        Is3 = get_field('Is3')
        df['Is1'] = Is3[:, 0]
        df['Is2'] = Is3[:, 1]
        df['Is3'] = Is3[:, 2]
        df['Is'] = np.mean(Is3, axis=1)
        # [mA]
        df['Iab'] = df['Is'] * 1000
        Il3 = get_field('Il3')
        df['Il1'] = Il3[:, 0]
        df['Il2'] = Il3[:, 1]
        df['Il3'] = Il3[:, 2]
        df['Il'] = np.mean(Il3, axis=1)
        # [mA]
        df['Ileakage'] = df['Il'] * 1000

        df['frequency'] = frequencies[f_id]
        dfl.append(df)

    df = pd.concat(dfl)

    return df


def _extract_emd(mat, **kwargs):
    emd = mat['EMD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, emd.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = emd[f_id]
        # fdata_md = md[f_id]

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in fdata['Time'].squeeze()]
        ).T
        df = pd.DataFrame(
            np.hstack((
                timestamp,
                fdata['ni'],
                fdata['nu'][:, np.newaxis],
                fdata['Z3'],
                fdata['Is3'],
                fdata['Il3'],
                fdata['Zg3'],
            )),
        )
        df.columns = (
            'datetime',
            'a',
            'b',
            'p',
            'Z1',
            'Z2',
            'Z3',
            'Is1',
            'Is2',
            'Is3',
            'Il1',
            'Il2',
            'Il3',
            'Zg1',
            'Zg2',
            'Zg3',
        )

        df['frequency'] = np.ones(df.shape[0]) * fdata['fm']

        # cast to correct type
        df['datetime'] = pd.to_datetime(df['datetime'])
        df['a'] = df['a'].astype(int)
        df['b'] = df['b'].astype(int)
        df['p'] = df['p'].astype(int)

        df['Z1'] = df['Z1'].astype(complex)
        df['Z2'] = df['Z2'].astype(complex)
        df['Z3'] = df['Z3'].astype(complex)

        df['Zg1'] = df['Zg1'].astype(complex)
        df['Zg2'] = df['Zg2'].astype(complex)
        df['Zg3'] = df['Zg3'].astype(complex)

        df['Is1'] = df['Is1'].astype(complex)
        df['Is2'] = df['Is2'].astype(complex)
        df['Is3'] = df['Is3'].astype(complex)

        df['Il1'] = df['Il1'].astype(complex)
        df['Il2'] = df['Il2'].astype(complex)
        df['Il3'] = df['Il3'].astype(complex)

        dfl.append(df)

    if len(dfl) == 0:
        return None
    df = pd.concat(dfl)

    # average swapped current injections here!
    df = _average_swapped_current_injections(df)

    # sort current injections
    condition = df['a'] > df['b']
    df.loc[condition, ['a', 'b']] = df.loc[
        condition, ['b', 'a']
    ].values.astype(int)
    # for some reason we lose the integer casting of a and b here
    df['a'] = df['a'].astype(int)
    df['b'] = df['b'].astype(int)

    # change sign because we changed A and B
    df.loc[condition, ['Z1', 'Z2', 'Z3']] *= -1

    # average of Z1-Z3
    df['Zt'] = np.mean(df[['Z1', 'Z2', 'Z3']].values, axis=1)
    # we need to keep the sign of the real part
    sign_re = np.real(df['Zt']) / np.abs(np.real(df['Zt']))
    df['r'] = np.abs(df['Zt']) * sign_re
    # df['Zt_std'] = np.std(df[['Z1', 'Z2', 'Z3']].values, axis=1)

    df['Is'] = np.mean(df[['Is1', 'Is2', 'Is3']].values, axis=1)
    df['Il'] = np.mean(df[['Il1', 'Il2', 'Il3']].values, axis=1)
    df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']].values, axis=1)

    # "standard" injected current, in [mA]
    df['Iab'] = np.abs(df['Is']) * 1e3
    df['Iab'] = df['Iab'].astype(float)
    # df['Is_std'] = np.std(df[['Is1', 'Is2', 'Is3']].values, axis=1)
    # take absolute value and convert to mA
    df['Ileakage'] = np.abs(df['Il']) * 1e3
    df['Ileakage'] = df['Ileakage'].astype(float)

    return df



import functools


def enable_result_transforms(func):
    """Decorator that tries to use the object provided using a kwarg called
    'electrode_transformator' to transform the return values of an import
    function. It is intended to be used to transform electrode numbers and
    locations, i.e. for use in roll-along-measurement schemes.

    The transformator object must have a function .transform, which takes three
    parameters: data, electrode, topography and returns three correspondingly
    transformed objects.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        func_transformator = kwargs.pop('electrode_transformator', None)
        data, electrodes, topography = func(*args, **kwargs)
        if func_transformator is not None:
            data_transformed, electrodes_transformed, \
                topography_transformed = func_transformator.transform(
                    data, electrodes, topography
                )
            return data_transformed, electrodes_transformed, \
                topography_transformed
        else:
            return data, electrodes, topography
    return wrapper





#!/usr/bin/env python
import os
import glob

from setuptools import setup, find_packages

# under windows, run
# python.exe setup.py bdist --format msi
# to create a windows installer

# version_long = '0.1.6.dev0'
version_long = '0.1.6.1'

# package data
os.chdir('lib/reda/testing')
package_data = glob.glob('data/*')
os.chdir('../../../')


if __name__ == '__main__':
    setup(
        name='reda',
        version=version_long,
        description='Reproducible Electrical Data Analysis',
        long_description=open('README.md', 'r').read(),
        long_description_content_type="text/markdown",
        author='Maximilian Weigand and Florian M. Wagner',
        author_email='mweigand@geo.uni-bonn.de, wagner@geo.uni-bonn.de',
        license='MIT',
        url='https://github.com/geophysics-ubonn/reda',
        packages=find_packages("lib"),
        package_dir={'': 'lib'},
        package_data={'reda.testing': package_data},
        install_requires=[
            'numpy',
            'scipy',
            'pandas',
            'matplotlib',
            'pillow',
            'h5py',
        ],
        extras_require={
            "crtomo":  [
                "crtomo_tools",
            ],
        },
        classifiers=(
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ),
    )



"""Import data from the EIT-systems built at the Research Center Jülich (FZJ).

As there is an increasing number of slightly different file formats in use,
this module acts as an selector for the appropriate import functions.
"""
import functools
import os
# import logging

import numpy as np
import pandas as pd
import scipy.io as sio

import reda.importers.eit_version_2010 as eit_version_2010
import reda.importers.eit_version_2013 as eit_version_2013
import reda.importers.eit_version_2017 as eit_version_2017
import reda.importers.eit_version_2018a as eit_version_2018a
import reda.importers.eit_version_20200609 as eit_version_20200609
from reda.importers.fzj_readbin import fzj_readbin

from reda.importers.utils.decorators import enable_result_transforms

from reda.configs.configManager import ConfigManager

# data file formats differ slightly between versions. Version numbers do not
# follow a consistent naming scheme. Therefore we introduce this dict to map
# the version numbers found in the .mat files to the reda python modules.
mat_version_importers = {
    # this is the file version used for the 160 channel multiplexer system
    'FZJ-EZ-2018A': eit_version_2018a,
    'FZJ-EZ-2017': eit_version_2017,
    'FZJ-EZ-09.11.2010': eit_version_2010,
    'FZJ-EZ-14.02.2013': eit_version_2013,
    'EZ-2020-06-09': eit_version_20200609,
}


def _get_file_version(filename):
    """High level import function that tries to determine the specific version
    of the data format used.

    Parameters
    ----------
    filename: string
        File path to a .mat matlab filename, as produced by the various
        versions of the emmt_pp.exe postprocessing program.

    Returns
    -------
    version: string
        a sanitized version of the file format version

    """
    mat = sio.loadmat(filename, squeeze_me=True)
    version = mat['MP']['Version'].item()
    del(mat)

    return version


def MD_ConfigsPermutate(df_md):
    """Given a MD DataFrame, return a Nx4 array which permutes the current
    injection dipoles.
    """
    g_current_injections = df_md.groupby(['a', 'b'])
    ab = np.array(list(g_current_injections.groups.keys()))
    config_mgr = ConfigManager(nr_of_electrodes=ab.max())
    config_mgr.gen_configs_permutate(ab, silent=True)
    return config_mgr.configs


def get_mnu0_data(filename, configs, return_3p=False, **kwargs):
    """Import data post-processed as 3P data (NMU0), i.e., measured towards
    common ground.

    Parameters
    ----------
    filename : string (usually: eit_data_mnu0.mat)
        filename of matlab file
    configs : Nx4 numpy.ndarray|filename|function
        4P measurements configurations (ABMN) to generate out of the data. If
        this parameter is a callable, then call it with the MD DataFrame as its
        sole parameter and expect a Nx4 numpy.ndarray as return value
    return_3p : bool, optional
        also return 3P data

    Keyword Arguments
    -----------------
    multiplexer_group : int|None, optional
        For the multiplexer system (version 2018a) the multiplexer group MUST
        be specified to import data. This is a number between 1 and 4.
    compute_errors : None|str, optional
        If this parameter points to the .bin file containing the raw data, then
        compute data errors based on a noise level analysis of individual time
        series. Time-consuming!

    Returns
    -------
    data_emd_4p : pandas.DataFrame
        The generated 4P data
    data_md_raw : pandas.DataFrame|None
        MD data (sometimes this data is not imported, then we return None here)
    data_emd_3p : pandas.DataFrame
        The imported 3P data (only if return_3p==True)
    """
    if not os.path.isfile(filename):
        raise IOError('Data file not found! {}'.format(filename))

    version = _get_file_version(filename)
    importer = mat_version_importers.get(version, None)
    if importer is not None:
        mat = sio.loadmat(filename, squeeze_me=True)
        data_md_raw = importer._extract_md(mat, **kwargs)
        data_emd_3p = importer._extract_emd(mat, **kwargs)

        # check configs
        if callable(configs):
            configs_abmn = configs(data_md_raw)
        else:
            configs_abmn = configs

        if data_emd_3p is not None:
            data_emd_4p = compute_quadrupoles(
                data_emd_3p, configs_abmn, data_md_raw)
        else:
            data_emd_4p = None

        binary_file = kwargs.get('compute_errors', None)
        if binary_file is not None:
            assert os.path.isfile(
                binary_file), 'compute_errors must point to a valid .bin file'

            adc_data = importer._extract_adc_data(mat, **kwargs)
            data_emd_4p = compute_data_errors(
                data_emd_4p,
                data_md_raw,
                adc_data,
                binary_file,
                **kwargs,
            )
    else:
        raise Exception(
            'The file version "{}" is not supported yet.'.format(
                version)
        )

    if return_3p:
        return data_emd_4p, data_md_raw, data_emd_3p
    else:
        return data_emd_4p, data_md_raw


def get_md_data(filename, **kwargs):
    """Import data and return the MD (i.e., injection) data

    Parameters
    ----------
    filename : string (usually: eit_data_mnu0.mat)
        filename of matlab file

    Keyword Arguments
    -----------------
    multiplexer_group : int|None, optional
        For the multiplexer system (version 2018a) the multiplexer group MUST
        be specified to import data. This is a number between 1 and 4.

    Returns
    -------
    data_md_raw : pandas.DataFrame|None
        MD data (sometimes this data is not imported, then we return None here)
    """
    if not os.path.isfile(filename):
        raise IOError('Data file not found! {}'.format(filename))

    version = _get_file_version(filename)
    importer = mat_version_importers.get(version, None)
    if importer is not None:
        mat = sio.loadmat(filename, squeeze_me=True)
        data_md_raw = importer._extract_md(mat, **kwargs)
        return data_md_raw
    else:
        raise Exception('emmt_pp version not found: {}'.format(version))


def get_adc_data(filename, **kwargs):
    """Import data and return the adc-related data from the MD (i.e.,
    injection) structure

    Parameters
    ----------
    filename : string (usually: eit_data_mnu0.mat)
        filename of matlab file

    Keyword Arguments
    -----------------
    multiplexer_group : int|None, optional
        For the multiplexer system (version 2018a) the multiplexer group MUST
        be specified to import data. This is a number between 1 and 4.

    Returns
    -------
    data_adc_raw : pandas.DataFrame|None
        adc-MD data (sometimes this data is not imported, then we return None
        here)
    """
    if not os.path.isfile(filename):
        raise IOError('Data file not found! {}'.format(filename))

    version = _get_file_version(filename)
    importer = mat_version_importers.get(version, None)
    if importer is not None:
        mat = sio.loadmat(filename, squeeze_me=True)
        data_md_raw = importer._extract_adc_data(mat, **kwargs)
        return data_md_raw
    else:
        raise Exception('emmt_pp version not found: {}'.format(version))


@enable_result_transforms
@functools.wraps(get_mnu0_data)
def read_3p_data(*args, **kwargs):
    # this is a wrapper that conforms to the importer standards
    results = get_mnu0_data(*args, **kwargs)
    df_emd = results[0]
    return df_emd, None, None


def compute_quadrupoles(df_emd, config_file, df_md=None):
    """
    Parameters
    ----------
    df_emd : pandas.DataFrame
        The EMD data, as imported from the .mat file (3P-data)
    config_file : string
        filename for configuration file. The configuration file contains N rows
        with 4 columns each (A, B, M, N)
    df_md : pandas.DataFrame (optional)
        The MD data

    Returns
    -------

    """
    # 'configs' can be a numpy array or a filename
    if not isinstance(config_file, np.ndarray):
        configs = np.loadtxt(config_file).astype(int)
    else:
        configs = config_file

    configs = np.atleast_2d(configs)

    # construct four-point measurements via superposition
    print('Constructing four-point measurements')
    quadpole_list = []
    index = 0
    for Ar, Br, M, N in configs:
        # print('constructing', Ar, Br, M, N)
        # the order of A and B doesn't concern us
        A = np.min((Ar, Br))
        B = np.max((Ar, Br))

        # first choice: correct ordering
        query_M = df_emd.query('a=={0} and b=={1} and p=={2}'.format(
            A, B, M
        )).sort_values('datetime')
        query_N = df_emd.query('a=={0} and b=={1} and p=={2}'.format(
            A, B, N
        )).sort_values('datetime')

        if query_M.size == 0 or query_N.size == 0:
            continue

        if query_M.size != query_N.size:
            raise Exception(
                'There is something wrong with the data, different sized ' +
                'data sets for M and N'
            )

        index += 1

        # keep these columns as they are (no subtracting)
        keep_cols_all = [
            'datetime',
            'frequency',
            'a', 'b',
            'Zg1', 'Zg2', 'Zg3',
            'Zg',
            'Is',
            'Il',
            'Iab',
            'Ileakage',
        ]
        # only keep those are actually there
        keep_cols = [x for x in keep_cols_all if x in query_M.columns]

        df4 = pd.DataFrame()
        diff_cols = ['Zt', ]
        df4[keep_cols] = query_M[keep_cols]
        for col in diff_cols:
            df4[col] = query_M[col].values - query_N[col].values
        df4['m'] = query_M['p'].values
        df4['n'] = query_N['p'].values

        quadpole_list.append(df4)

    if quadpole_list:
        dfn = pd.concat(quadpole_list)
        Rsign = np.sign(np.real(dfn['Zt']))
        dfn['r'] = Rsign * np.abs(dfn['Zt'])
        if 'Iab' in dfn.columns:
            dfn['Vmn'] = dfn['r'] * dfn['Iab']
        dfn['rpha'] = np.arctan2(
            np.imag(dfn['Zt'].values),
            np.real(dfn['Zt'].values)
        ) * 1e3
        # Depending on the specific analysis software ware, some columns are
        # located in the md struct and need to be merged to the dfn
        check_md_columns = [
            'Zg',
            'Iab',
            'Ileakage',
        ]
        for column in check_md_columns:
            if(column not in dfn.columns and df_md is not None and
                    column in df_md.columns):
                print('Adding column {} from MD'.format(column))
                # import IPython
                # IPython.embed()
                dfn = pd.merge(
                    dfn,
                    df_md[['a', 'b', 'frequency', 'datetime', column]],
                    on=['a', 'b', 'frequency', 'datetime'],
                    how='left',
                )
    else:
        dfn = pd.DataFrame()

    return dfn.sort_values(['frequency', 'a', 'b', 'm', 'n'])


def apply_correction_factors(df, correction_data):
    """Apply correction factors for a pseudo-2D measurement setup. See Weigand
    and Kemna, 2017, Biogeosciences, for detailed information.

    Parameters
    ----------
    df : :py:class:`pandas.DataFrame`
        DataFrame containing the data
    correction_data : string|iterable of strings|:py:class:`numpy.ndarray`
        Correction data, either as a filename, a list of filenames to be
        merged, or directly as a numpy array

    Returns
    -------
    df : :py:class:`pandas.DataFrame`
        Corrected data
    corr_data : :py:class:`numpy.ndarray`
        Correction factors used

    """
    if isinstance(correction_data, (list, tuple)):
        corr_data_raw = np.vstack(
            [np.loadtxt(x) for x in correction_data]
        )
    elif isinstance(correction_data, np.ndarray):
        corr_data_raw = correction_data
    else:
        # assume only one data file
        corr_data_raw = np.loadtxt(correction_data)

    assert corr_data_raw.shape[1] in (3, 5)
    # if required, convert from CRTomo electrode denotations in (a,b,m,n) style
    if corr_data_raw.shape[1] == 3:
        A = (corr_data_raw[:, 0] / 1e4).astype(int)
        B = (corr_data_raw[:, 0] % 1e4).astype(int)
        M = (corr_data_raw[:, 1] / 1e4).astype(int)
        N = (corr_data_raw[:, 1] % 1e4).astype(int)
        corr_data = np.vstack((A, B, M, N, corr_data_raw[:, 2])).T
    else:
        corr_data = corr_data_raw

    corr_data[:, 0:2] = np.sort(corr_data[:, 0:2], axis=1)
    corr_data[:, 2:4] = np.sort(corr_data[:, 2:4], axis=1)

    # if 'frequency' not in df.columns:
    #     raise Exception(
    #         'No frequency data found. Are you sure this is a seit data set?'
    #     )

    df = df.reset_index()
    gf = df.groupby(['a', 'b', 'm', 'n'])
    for key, item in gf.indices.items():
        # print('key', key)
        # print(item)
        item_norm = np.hstack((np.sort(key[0:2]), np.sort(key[2:4])))
        # print(item_norm)
        index = np.where(
            (corr_data[:, 0] == item_norm[0]) &
            (corr_data[:, 1] == item_norm[1]) &
            (corr_data[:, 2] == item_norm[2]) &
            (corr_data[:, 3] == item_norm[3])
        )[0]
        # print(index, corr_data[index])
        if len(index) == 0:
            print(key)
            raise Exception(
                'No correction factor found for this configuration'
            )

        factor = np.ones_like(item) * corr_data[index, 4]
        # apply correction factor
        for col in ('r', 'Zt', 'Vmn', 'rho_a'):
            if col in df.columns:
                df.iloc[item, df.columns.get_loc(col)] *= factor
        # add the correction factor to the DataFrame
        if 'corr_fac' not in df.columns:
            df['corr_fac'] = np.nan
        df.iloc[item, df.columns.get_loc('corr_fac')] = factor
    return df, corr_data


def compute_data_errors(
        data_emd_4p, data_md_raw, adc_data, binary_file, **kwargs):
    """Compute data errors based on a noise-level analysis and subsequent
    linear error propagation.

    Additional Parameters
    ---------------------
    error_model_use_diff_noise_level : bool, optional (default: False)

    What we need in data_emd_4p:

        * phi_m
        * phi_n
        * dphi_m
        * dphi_n
        * current (complex)

    """
    print('Preparing data_emd_4p')
    # print('Merging in complex current')
    # import IPython
    # IPython.embed()

    if 'Is' not in data_emd_4p.columns:
        data_emd_4p = pd.merge(
            data_emd_4p,
            data_md_raw[['a', 'b', 'frequency', 'datetime', 'Is']],
            on=['a', 'b', 'frequency', 'datetime'],
            how='left',
        )

    print('Add electrode potentials')
    us3 = adc_data.xs('Us', axis=1, level=1)
    us3_selected = pd.merge(
        data_emd_4p[['a', 'b', 'frequency', 'datetime']],
        us3,
        on=['a', 'b', 'frequency', 'datetime'],
        how='left',
    ).set_index(['a', 'b', 'frequency', 'datetime'])
    data_emd_4p['pot_m'] = us3_selected.values[
        np.array(range(us3_selected.shape[0])), data_emd_4p['m'].values - 1]
    data_emd_4p['pot_n'] = us3_selected.values[
        np.array(range(us3_selected.shape[0])), data_emd_4p['n'].values - 1]

    obj = fzj_readbin(binary_file)

    # check = (data_emd_4p['pot_m'] - data_emd_4p['pot_n']) / data_emd_4p['Is']
    print('Compute noise levels')

    def get_noise_levels(row):
        indices = obj.find_swapped_measurement_indices(
            row['a'],
            row['b'],
            row['frequency'],
            row['datetime']
        )
        noise_levels = []
        for index in indices:
            # try to remove systematic components
            ts_m = obj.data[index][row['m'] - 1, :]
            ts_n = obj.data[index][row['n'] - 1, :]
            ts_diff = (ts_m - ts_m.mean()) - (ts_n - ts_n.mean())

            fdata = obj.frequency_data.iloc[index]
            fs = fdata['sampling_frequency'] / fdata['oversampling']

            fft, u_peaks, noise_level = obj._get_noise_level_from_fft(
                ts_diff,
                fs=fs,
            )

            # now analyze the channels separately
            level1 = obj.fft_analysis_one_channel(
                index,
                row['m'],
                remove_excitation_frequency=kwargs.get(
                    'remove_excitation_frequency', False
                ),
                remove_noise_harmonics=kwargs.get(
                    'remove_noise_harmonics', False
                ),
            )[0]
            level2 = obj.fft_analysis_one_channel(
                index,
                row['n'],
                remove_excitation_frequency=kwargs.get(
                    'remove_excitation_frequency', False
                ),
                remove_noise_harmonics=kwargs.get(
                    'remove_noise_harmonics', False
                ),
            )[0]

            # noise level of difference of both time series
            # logger = logging.getLogger()
            # logger.warning(
            #     "IMPORTANT: I'm using the noise level of the ts-difference!")
            # we assume that the resulting noise component is the result of
            # both equally sized noise components of the single time series:
            # diff = t1 - t2
            # -> d_diff = sqrt(dt1** 2 + dt2 ** 2)
            # if we assume dt1 == dt2, then
            # dt1 = dt2 = 1 / np.sqrt(2) d_diff
            if kwargs.get('error_model_use_diff_noise_level', False):
                noise_levels.append(noise_level / np.sqrt(2))
                noise_levels.append(noise_level / np.sqrt(2))
            else:
                # individual channel noise levels
                noise_levels.append(level1)
                noise_levels.append(level2)

            # current channel noise analysis
            # TODO: Do properly, for now we only analyse channel 41
            ts_current = obj.data[index][41, :]

            fft, u_peaks, noise_level_current = obj._get_noise_level_from_fft(
                ts_current - ts_current.mean(),
                fs=fs,
            )
            noise_levels.append(noise_level_current / 1000)

        return np.array(noise_levels)

    noise_levels = data_emd_4p[
        ['a', 'b', 'frequency', 'datetime', 'm', 'n']
    ].apply(get_noise_levels, axis=1).values

    noise_levels = np.concatenate(noise_levels, axis=1).T

    # hack: convert to voltages: FFT <-> voltage noise level
    noise_levels /= 48

    # propagate noise level on time -series down through the lockin
    noise_levels /= 40

    # import IPython
    # IPython.embed()

    # regular injection
    data_emd_4p['dpot_m_1'] = noise_levels[:, 0]
    data_emd_4p['dpot_n_1'] = noise_levels[:, 1]
    data_emd_4p['dcurrent_1'] = noise_levels[:, 2]
    # swapped injection
    data_emd_4p['dpot_m_2'] = noise_levels[:, 3]
    data_emd_4p['dpot_n_2'] = noise_levels[:, 4]
    data_emd_4p['dcurrent_2'] = noise_levels[:, 5]

    data_emd_4p['dpot_m'] = np.sqrt(
        data_emd_4p['dpot_m_1'] ** 2 / 2 +
        data_emd_4p['dpot_m_2'] ** 2 / 2
    )
    data_emd_4p['dpot_n'] = np.sqrt(
        data_emd_4p['dpot_n_1'] ** 2 / 2 +
        data_emd_4p['dpot_n_2'] ** 2 / 2
    )
    data_emd_4p['dcurrent'] = np.sqrt(
        data_emd_4p['dcurrent_1'] ** 2 / 2 +
        data_emd_4p['dcurrent_2'] ** 2 / 2
    )

    magnitude_errors = compute_magnitude_errors(
        data_emd_4p['pot_m'].values,
        data_emd_4p['pot_n'].values,
        data_emd_4p['Is'].values,
        # test: 1 / sqrt(3) to simulate the three repetitions
        data_emd_4p['dpot_m'].values,  # / np.sqrt(3),
        data_emd_4p['dpot_n'].values,  # / np.sqrt(3),
        dcurrent=data_emd_4p['dcurrent'].values,
        # dcurrent=np.array(0),
        **kwargs
    )

    phase_errors = compute_phase_errors(
        data_emd_4p['pot_m'].values,
        data_emd_4p['pot_n'].values,
        data_emd_4p['Is'].values,
        # test: 1 / sqrt(3) to simulate the three repetitions
        data_emd_4p['dpot_m'].values,  # / np.sqrt(3),
        data_emd_4p['dpot_n'].values,  # / np.sqrt(3),
        dcurrent=data_emd_4p['dcurrent'].values,
        # dcurrent=np.array(0),
        **kwargs
    )

    data_emd_4p['r_error'] = magnitude_errors
    data_emd_4p['rpha_error'] = phase_errors * 1000

    # import IPython
    # IPython.embed()
    return data_emd_4p


def compute_magnitude_errors(
        phi_m, phi_n, current, dphi_m, dphi_n, dcurrent, **kwargs):
    """Compute magnitude errors based on linear error propagation
    """
    error_m = np.abs(dphi_m) / np.abs(current)
    error_n = np.abs(dphi_n) / np.abs(current)
    error_current = -np.abs(
        phi_m - phi_n
    ) / current ** 2 * dcurrent

    error_magnitude = np.sqrt(
        error_m ** 2 +
        error_n ** 2 +
        error_current ** 2 +
        0
    )
    return np.abs(error_magnitude)


def compute_phase_errors(
        phi_m, phi_n, current, dphi_m, dphi_n, dcurrent, **kwargs):
    """Compute the phase error based on linear error propagation of the
    transfer impedance equation:

        Zt = U_mn / I_ab = (phi_m - phi_n) / I_ab

    Parameters
    ----------
    phi_m : numpy.ndarray|complex float
        Complex potential at electrode m
    phi_n : numpy.ndarray|complex float
        Complex potential at electrode n
    current : numpy.ndarray|complex float
        Complex current injected at electrodes a and b
    dphi_m : numpy.ndarray|complex float
        potential error at electrode m. Real part corresponds to error of the
        real part of the potential, vice versa for the imaginary part
    dphi_n : numpy.ndarray|complex float
        potential error at electrode m. Real part corresponds to error of the
        real part of the potential, vice versa for the imaginary part
    dcurrent : numpy.ndarray|complex float
        current error at electrode m. Real part corresponds to error of the
        real part of the potential, vice versa for the imaginary part

    Returns
    -------
    error_phase : numpy.ndarray|float
        Phase error in [rad]
    """

    A = - current.imag * (
        phi_m.real - phi_n.real
    ) + current.real * (phi_m.imag - phi_n.imag)

    B = current.real * (phi_m.real - phi_n.real) + current.imag * (
        phi_m.imag - phi_n.imag
    )

    # chain rule: d/dx arctan(x) = 1 / (1 + x^2)
    prefix = 1 / (1 + (A / B) ** 2)

    error_dphi_m_real = prefix * (
        -current.imag / B + A / B ** 2 * current.real
    ) * dphi_m

    error_dphi_n_real = prefix * (
        current.imag / B - A / B ** 2 * current.real
    ) * dphi_m

    error_dphi_m_imag = prefix * (
        current.real / B + A / B ** 2 * current.imag
    ) * dphi_m

    error_dphi_n_imag = prefix * (
        -current.real / B - A / B ** 2 * current.imag
    ) * dphi_n

    error_current_real = prefix * (
        (phi_m.imag - phi_n.imag) / B + A / B ** 2 * (phi_m.real - phi_n.real)
    ) * dcurrent

    error_current_imag = prefix * (
        -1 * (phi_m.real - phi_n.real) / B +
        A / B ** 2 * (phi_m.imag - phi_n.imag)
    ) * dcurrent

    inner_term = 0
    inner_term += error_dphi_m_real ** 2
    inner_term += error_dphi_n_real ** 2
    inner_term += error_dphi_m_imag ** 2
    inner_term += error_dphi_n_imag ** 2
    if not kwargs.get('errmod_pha_no_current_error', False):
        inner_term += error_current_real ** 2
        inner_term += error_current_imag ** 2

    phase_error = np.sqrt(inner_term)

    return phase_error



import logging
import functools
import os

import pandas as pd

from reda.utils.electrode_manager import electrode_manager
from reda.containers.BaseContainer import ImportersBase
from reda.containers.BaseContainer import BaseContainer

import reda.importers.bert as reda_bert_import
import reda.importers.iris_syscal_pro as reda_syscal
import reda.importers.mpt_das1 as reda_mpt
from reda.importers.crtomo import load_mod_file
from reda.importers.tsert_import import tsert_import
from reda.exporters.tsert_export import tsert_export

from reda.utils.norrec import assign_norrec_to_df, average_repetitions

from reda.utils.decorators_and_managers import append_doc_of
from reda.utils.decorators_and_managers import LogDataChanges

logger = logging.getLogger(__name__)


class ERTImporters(ImportersBase):
    """This class provides wrappers for most of the importer functions and is
    meant to be inherited by the ERT data container.

    See Also
    --------
    Exporters
    """
    def tsert_summary(self, filename, **kwargs):
        """Try to open a given filename (usually a .h5 file) as a TSERT file
        and print out a summary of contained data.

        Parameters
        ----------
        filename : str
            Filename of data file
        """
        importer = tsert_import(filename)
        importer.summary(**kwargs)

    def import_tsert(
            self, filename, timesteps='all', version='base', **kwargs):
        """TSERT import

        Parameters
        ----------
        filename : str
            Path to hdf file to import data from
        timesteps : str|list|datetime.datetime
            Timesteps that should be imported
        version : str
            Which version of the data to load. Time steps that do not have this
            specific version are ignored. Default: base version
        """

        self.logger.info('TSERT import')
        importer = tsert_import(filename)
        with LogDataChanges(self, filter_action='import'):
            data = importer.import_data(
                timesteps=timesteps,
                version=version,
                **kwargs,
            )
            electrode_positions_df = importer.load_electrode_positions()
            electrode_positions = electrode_manager(electrode_positions_df)
            topography = importer.load_topography()
            metadata = importer.load_metadata()

            self._add_to_container(
                data, electrode_positions, topography, metadata
            )
        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    def import_crtomo_data(self, filename, **kwargs):
        """
        Import a CRTomo-style measurement file (usually: volt.dat).

        Parameters
        ----------
        filename : str
            path to data file
        """
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])

        with LogDataChanges(self, filter_action='import'):
            data = load_mod_file(filename)
            if timestep is not None:
                data['timestep'] = timestep
            self._add_to_container(data)

        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    @append_doc_of(reda_syscal.import_bin)
    def import_syscal_bin(self, filename, **kwargs):
        """Syscal import

        timestep: int or :class:`datetime.datetime`
            if provided use this value to set the 'timestep' column of the
            produced dataframe. Default: 0

        """
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])
        self.logger.info('IRIS Syscal Pro bin import')
        with LogDataChanges(self, filter_action='import'):
            data, electrode_positions, topography = reda_syscal.import_bin(
                filename, **kwargs
            )
            if timestep is not None:
                data['timestep'] = timestep
            self._add_to_container(data, electrode_positions, topography)
        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    @append_doc_of(reda_syscal.import_txt)
    def import_syscal_txt(self, filename, **kwargs):

        """
        Syscal import

        timestep: int or :class:`datetime.datetime`
            if provided use this value to set the 'timestep' column of the
            produced dataframe. Default: 0

        """

        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])
        self.logger.info('IRIS Syscal Pro text import')
        with LogDataChanges(self, filter_action='import'):
            data, electrode_positions, topography = reda_syscal.import_txt(
                filename, **kwargs)
            if timestep is not None:
                data['timestep'] = timestep
            self._add_to_container(data, electrode_positions, topography)
        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    @append_doc_of(reda_bert_import.import_ohm)
    def import_bert(self, filename, **kwargs):
        """BERT .ohm file import"""
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])

        self.logger.info('Unified data format (BERT/pyGIMLi) file import')
        with LogDataChanges(self, filter_action='import',
                            filter_query=os.path.basename(filename)):
            data, electrodes, topography = reda_bert_import.import_ohm(
                filename, **kwargs)
            if timestep is not None:
                data['timestep'] = timestep
            self._add_to_container(data)
            self.electrode_positions = electrodes  # See issue #22
        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    @append_doc_of(reda_mpt.import_das1)
    def import_mpt(self, filename, **kwargs):
        """MPT DAS 1 importer

        timestep : int or :class:`datetime.datetime`
            if provided use this value to set the 'timestep' column of the
            produced dataframe. Default: 0

        """
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])
        self.logger.info('MPT DAS-1 import')
        with LogDataChanges(self, filter_action='import'):
            data, electrodes, topography = reda_mpt.import_das1(
                filename, **kwargs)
            if timestep is not None:
                data['timestep'] = timestep
            # check if SIP data set is imported, if so select the lowest
            # possible frequency
            if 'frequency' in data.columns:
                data = data.query(
                    'frequency == {}'.format(data.frequency.min()))
                print(data)
            self._add_to_container(
                data[['a', 'b', 'm', 'n', 'r', 'dr', 'I', 'datetime']]
            )

        if kwargs.get('verbose', False):
            print('Summary:')
            self._describe_data(data)

    @functools.wraps(import_bert)
    def import_pygimli(self, *args, **kargs):
        self.import_bert(*args, **kargs)


class ERTExporters(object):
    def export_tsert(self, filename, version, **kwargs):
        """Export data to TSERT

        """
        exporter = tsert_export(filename)
        exporter.set_electrode_positions(self.electrode_positions)
        exporter.set_topography(self.topography)
        exporter.add_data(self.data, version, **kwargs)
        exporter.add_metadata(self.metadata)

    def export_to_pygimli_scheme(self, norrec='nor', timestep=None):
        """Export the data into a pygimili.DataContainerERT object.

        For now, do NOT set any sensor positions

        Parameters
        ----------


        Returns
        -------
        """
        logger.info('Exporting to pygimli DataContainer')
        logger.info('{} data will be exported'.format(norrec))
        if timestep is None:
            logger.info('No timestep selection is applied')
        else:
            logger.info('timestep(s) {} will be used'.format(timestep))

        import pygimli as pg
        data_container = pg.DataContainerERT()

        query = ' '.join((
            'norrec == "{}"'.format(norrec),
        ))

        if timestep is not None:
            query += ' and timestep=="{}"'.format(timestep)

        logger.debug('Query: {}'.format(query))

        subdata = self.data.query(query)
        assert subdata.shape[0] != 0

        data_container['a'] = subdata['a']
        data_container['b'] = subdata['b']
        data_container['m'] = subdata['m']
        data_container['n'] = subdata['n']
        data_container['r'] = subdata['r']

        if 'k' in subdata.columns:
            data_container['k'] = subdata['k']

        if 'rho_a' in subdata.columns:
            data_container['rhoa'] = subdata['rho_a']

        return data_container


class ERT(BaseContainer, ERTImporters, ERTExporters):
    """."""

    def __init__(
            self, data=None, electrode_positions=None, topography=None,
            metadata=None, **kwargs):
        """
        Parameters
        ----------
        data : :py:class:`pandas.DataFrame`
            If not None, then the provided DataFrame is assumed to contain
            valid data previously prepared elsewhere. Please refer to the
            documentation for required columns.
        electrode_positions : :py:class:`pandas.DataFrame`
            If set, this is expected to be a DataFrame which contains electrode
            positions with columns: "x", "y", "z".
        topography : :py:class:`pandas.DataFrame`
            If set, this is expected to a DataFrame which contains topography
            information with columns: "x", "y", "z".

        """
        self.required_columns = [
            'a',
            'b',
            'm',
            'n',
            'r',
        ]
        super().__init__(
            data,
            electrode_positions,
            topography,
            metadata,
            **kwargs
        )

    def check_dataframe(self, dataframe):
        """Check the given dataframe for the required type and columns
        """
        if dataframe is None:
            return None

        # is this a DataFrame
        if not isinstance(dataframe, pd.DataFrame):
            raise Exception(
                'The provided dataframe object is not a pandas.DataFrame')

        for column in self.required_columns:
            if column not in dataframe:
                raise Exception(
                    'Required column not in dataframe: {0}'.format(column))
        return dataframe

    def compute_reciprocal_errors(self, key="r"):
        # JG: BaseContainer?
        r"""
        Compute reciprocal erros following LaBrecque et al. (1996) according
        to:

        .. math::

            \epsilon = \left|\frac{2(|R_n| - |R_r|)}{|R_n| + |R_r|}\right|

        Parameters
        ----------
        key : str
            Parameter to calculate the reciprocal error for (default is "r").

        Examples
        --------
        >>> import reda
        >>> ert = reda.ERT()
        >>> ert.data = reda.utils.norrec.get_test_df()
        >>> ert.data = pd.DataFrame([
        ...     [1,2,3,4,95],
        ...     [3,4,2,1,-105]], columns=list("abmnr")
        ... )
        >>> ert.compute_reciprocal_errors()
        >>> ert.data["error"].mean() == 0.1
        True
        """

        # Assign norrec ids if not already present
        if "id" not in self.data.keys():
            self.data = assign_norrec_to_df(self.data)

        # Average repetitions
        data = average_repetitions(self.data, "r")

        # Get configurations with reciprocals
        data = data.groupby("id").filter(lambda b: not b.shape[0] == 1)
        n = self.data.shape[0] - data.shape[0]
        if n > 0:
            print("Could not find reciprocals for %d configurations" % n)

        # Calc reciprocal error
        grouped = data.groupby("id")

        def _error(group):
            R_n = group["r"].iloc[0]
            R_r = group["r"].iloc[1]
            return abs(2 * (abs(R_n) - abs(R_r)) / (abs(R_n) + abs(R_r)))

        error = grouped.apply(_error)
        error.name = "error"
        self.data = pd.merge(
            self.data,
            error.to_frame().reset_index(), how='outer',
            on='id'
        )



import os


class EnterDirectory(object):
    """This is a context manager that enters a given directory and returns to
    the initial current directory after finishing
    """
    def __init__(self, directory):
        self.pwd = os.getcwd()
        self.directory = os.path.abspath(str(directory))

    def __enter__(self):
        os.chdir(self.directory)

    def __exit__(self, dtype, value, traceback):
        os.chdir(self.pwd)


class CreateEnterDirectory(EnterDirectory):
    """This is a context manager that enters a given directory and returns to
    the initial current directory after finishing. If the target directory does
    not exist, create it.
    """
    def __enter__(self):
        os.makedirs(self.directory, exist_ok=True)
        os.chdir(self.directory)



# -*- coding: utf-8 -*-
""" Tests for converter functions

Run with

nosetests test_convert.py -s -v
"""
import pytest
# from nose.tools import *
import numpy as np
import reda.eis.convert as sip_convert
import numpy.testing
from_keys = sip_convert.from_converters.keys()


class TestClass_input_styles(object):
    """
    Test the three input styles:

        * 1D
        * 2D - one spectrum
        * 2D - multiple spectra
    """
    def precompute_values(self):
        rmag = np.array([49.999, 31.623, 24.992])
        rpha = np.array([-5.0000, -321.7506, -24.9948])
        Z = rmag * np.exp(1j * rpha / 1000)
        Y = 1 / Z
        rre = np.real(Z)
        rim = np.imag(Z)
        cre = np.real(Y)
        cim = np.imag(Y)
        cmag = np.abs(Y)
        cpha = np.arctan2(np.imag(Y), np.real(Y)) * 1000

        self.rmag = rmag
        self.rpha = rpha
        self.rre = rre
        self.rim = rim
        self.rmim = -rim

        self.cmag = cmag
        self.cpha = cpha
        self.cre = cre
        self.cim = cim
        self.cmim = -cim

    def setup(self):
        self.precompute_values()

    def test_input_styles(self):
        # prepare test styles
        data_1d = np.hstack((self.cmag, self.cpha))
        data_2d_one_spec = np.vstack((self.cmag, self.cpha))
        data_2d_multi_specs = np.vstack((data_1d, data_1d))

        for data, one_spec in zip(
            (data_1d, data_2d_one_spec, data_2d_multi_specs),
                (False, True, False)):
            data_converted = sip_convert.convert('cmag_cpha', 'rmag_rpha',
                                                 data, one_spec)
            assert data.shape == data_converted.shape
            data_backconverted = sip_convert.convert('rmag_rpha', 'cmag_cpha',
                                                     data_converted, one_spec)
            numpy.testing.assert_almost_equal(
                data, data_backconverted,
                decimal=4)


class TestClass_test_converters():
    @classmethod
    def teardown(self):
        pass

    def precompute_values(self):
        rmag = np.array([49.999, 31.623, 24.992])
        rpha = np.array([-5.0000, -321.7506, -24.9948])
        Z = rmag * np.exp(1j * rpha / 1000)
        Y = 1 / Z
        rre = np.real(Z)
        rim = np.imag(Z)
        cre = np.real(Y)
        cim = np.imag(Y)
        cmag = np.abs(Y)
        cpha = np.arctan2(np.imag(Y), np.real(Y)) * 1000

        self.rmag = rmag
        self.rpha = rpha
        self.rre = rre
        self.rim = rim
        self.rmim = -rim

        self.cmag = cmag
        self.cpha = cpha
        self.cre = cre
        self.cim = cim
        self.cmim = -cim

    def setup(self):
        self.precompute_values()

    def check_from_function(self, func, input1, input2):
        input_data = np.hstack((input1, input2))
        out_cre, out_cim = func(input_data)
        output_data = np.hstack((out_cre, out_cim))
        true_result = np.hstack((self.cre, self.cim))
        diffs = np.abs(true_result - output_data.flatten())
        for term in diffs:
            # check to within 8 places
            assert term == pytest.approx(0, abs=1e-8)
            # assert_almost_equal(term, 0, 8)

    def check_to_function(self, func, output1, output2):
        output_data = func(self.cre, self.cim)
        true_result = np.hstack((output1, output2))
        diffs = np.abs(true_result - output_data.flatten())
        for term in diffs:
            # check to within 8 places
            # assert_almost_equal(term, 0, 8)
            assert term == pytest.approx(0, abs=1e-8)

    def test_from_cre_cim(self):
        self.check_from_function(sip_convert.from_cre_cim, self.cre, self.cim)

    def test_from_cre_cmim(self):
        self.check_from_function(sip_convert.from_cre_cmim, self.cre,
                                 self.cmim)

    def test_from_cmag_cpha(self):
        self.check_from_function(sip_convert.from_cmag_cpha, self.cmag,
                                 self.cpha)

    def test_from_rre_rim(self):
        self.check_from_function(sip_convert.from_rre_rim, self.rre, self.rim)

    def test_from_rre_rmim(self):
        self.check_from_function(sip_convert.from_rre_rmim, self.rre,
                                 self.rmim)

    def test_from_rmag_rpha(self):
        self.check_from_function(sip_convert.from_rmag_rpha, self.rmag,
                                 self.rpha)

    def test_to_cre_cim(self):
        self.check_to_function(sip_convert.to_cre_cim, self.cre, self.cim)

    def test_to_cre_mim(self):
        self.check_to_function(sip_convert.to_cre_cmim, self.cre, self.cmim)

    def test_to_cmag_cpha(self):
        self.check_to_function(sip_convert.to_cmag_cpha, self.cmag, self.cpha)

    def test_to_rre_rim(self):
        self.check_to_function(sip_convert.to_rre_rim, self.rre, self.rim)

    def test_convert(self):
        """
        We test by converting to all values, from all values. This is a chain
        test, i.e. we only provide the first input data, and then we use the
        output of this conversion to feed the next conversion. The last
        conversion then converts back to the initial data format, and thus we
        can compare those outputs to the hardcoded values.
        """
        from_keys = list(sorted(sip_convert.from_converters.keys()))

        # we know (require) that the first key is rmag_rpha
        initial_values = np.hstack((self.rmag, self.rpha))
        start_values = initial_values.copy()

        for nr, from_key in enumerate(from_keys):
            to = (nr + 1) % len(from_keys)
            output_data = sip_convert.convert(
                from_keys[nr], from_keys[to], start_values
            )
            start_values = output_data

        diffs = start_values - initial_values

        for term in diffs.flatten():
            # check to within 3 places
            # assert_almost_equal(term, 0, 3)
            assert term == pytest.approx(0, abs=1e-3)



# -*- coding: utf-8 -*-
# medusa data file format: FZJ-EZ-2017
import datetime
import pandas as pd
import numpy as np


def _extract_adc_data(mat, **kwargs):
    """Extract adc-channel related data (i.e., data that is captured for all 48
    channels of the 40-channel medusa system

    """
    md = mat['MD'].squeeze()
    frequencies = mat['MP']['fm'].take(0)

    # it seems that there exist different file formats under this same official
    # version.
    if md['fm'].size == frequencies.size:
        use_v = 0
    else:
        use_v = 1

    # print('@@@')
    # import IPython
    # IPython.embed()
    # exit()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, frequencies.size):
        frequency = frequencies[f_id]

        if use_v == 0:
            def get_field(key):
                return md[key][f_id]
        elif use_v == 1:
            def get_field(key):
                indices = np.where(
                    md['fm'].take(0) == frequencies[f_id])
                return md[key].take(0)[indices]

        # def get_field(key):
        #     indices = np.where(md['fm'].take(f_id) == frequencies[f_id])
        #     return md[key].take(f_id)[indices]

        # import IPython
        # IPython.embed()
        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in get_field('Time')]
        ).T.squeeze()
        # IPython.embed()
        # exit()
        Us3 = get_field('Us3')
        data_list = []
        for key, data in zip(
                (
                    'Us3_1',
                    'Us3_2',
                    'Us3_3',
                    'Us',
                ), (
                    Us3[:, :, 0],
                    Us3[:, :, 1],
                    Us3[:, :, 2],
                    np.mean(
                        Us3,
                        axis=2,
                    ),
                )):
            df = pd.DataFrame(
                data, columns=['ch{:02}'.format(i) for i in range(48)]).T
            df['parameter'] = key
            df.set_index('parameter', append=True, inplace=True)
            df = df.T
            # I THINK we can just sort a, b - the actual data should already
            # conform to the notation that the lower electrode number is
            # assigned to a
            df[['a', 'b']] = pd.DataFrame(np.sort(get_field('cni'), axis=1))
            df.set_index(['a', 'b'], inplace=True)
            data_list.append(df)

        df_all = data_list[0]
        if len(data_list) > 1:
            for subdata in data_list[1:]:
                df_all = pd.merge(
                    df_all, subdata, left_index=True, right_index=True
                )
        df_all['datetime'] = timestamp
        df_all['frequency'] = get_field('fm')
        dfl.append(df_all)

    dfl = pd.concat(dfl)
    dfl.set_index('frequency', append=True, inplace=True)
    dfl.sort_index(axis=0, inplace=True)
    dfl.sort_index(axis=1, inplace=True)

    dfl.columns.set_names('channel', level=0, inplace=True)
    dfl.set_index('datetime', append=True, inplace=True)

    return dfl
    # IPython.embed()
    # print('@@@@@@@@')
    # exit()

        # column_names = ['ch{:02}'.format(i) for i in range(48)]
        # ab = get_field('cni')

        # # Ug3
        # index_pairs = [
        #     (channel, 'Ug3_{}'.format(i)) for channel in column_names
        #     for i in range(3)
        # ]

        # Ug3 = get_field('Ug3')
        # ug3_reshaped = Ug3.reshape([Ug3.shape[0], Ug3.shape[1] * 3])
        # df = pd.DataFrame(
        #     ug3_reshaped,
        #     index=pd.MultiIndex.from_arrays(
        #         [
        #             ab[:, 0],
        #             ab[:, 1],
        #             np.ones(ab.shape[0]) * frequency,
        #             # timestamp
        #         ],
        #         names=[
        #             'a', 'b', 'frequency'
        #         ]
        #     ),
        #     columns=pd.MultiIndex.from_tuples(
        #        index_pairs, names=['channel', 'parameter'])
        # )
        # dfl.append(df)

        # # Ue3
        # index_pairs = [
        #     (channel, 'Ue3_{}'.format(i)) for channel in column_names
        #     for i in range(3)
        # ]

        # Ue3 = get_field('Ue3')
        # ue3_reshaped = Ue3.reshape([Ue3.shape[0], Ue3.shape[1] * 3])
        # df = pd.DataFrame(
        #     ue3_reshaped,
        #     index=pd.MultiIndex.from_arrays(
        #         [
        #             ab[:, 0],
        #             ab[:, 1],
        #             np.ones(ab.shape[0]) * frequency,
        #             # timestamp
        #         ],
        #         names=['a', 'b', 'frequency']
        #     ),
        #     columns=pd.MultiIndex.from_tuples(
        #        index_pairs, names=['channel', 'parameter'])
        # )
        # dfl.append(df)

        # # Us3
        # index_pairs = [
        #     (channel, 'Us3_{}'.format(i)) for channel in column_names
        #     for i in range(3)
        # ]

        # Us3 = get_field('Us3')
        # ue3_reshaped = Us3.reshape([Us3.shape[0], Us3.shape[1] * 3])
        # df = pd.DataFrame(
        #     ue3_reshaped,
        #     index=pd.MultiIndex.from_arrays(
        #         [
        #             ab[:, 0],
        #             ab[:, 1],
        #             np.ones(ab.shape[0]) * frequency,
        #             # timestamp
        #         ],
        #         names=['a', 'b', 'frequency']
        #     ),
        #     columns=pd.MultiIndex.from_tuples(
        #        index_pairs, names=['channel', 'parameter'])
        # )
        # dfl.append(df)

    # dfl = pd.concat(dfl)
    # dfl.sort_index(axis=0, inplace=True)
    # dfl.sort_index(axis=1, inplace=True)

    return dfl


def _extract_md(mat, **kwargs):
    md = mat['MD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, md.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = md[f_id]
        # for name in fdata.dtype.names:
        #     print(name, fdata[name].shape)

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in fdata['Time'].squeeze()]
        ).T
        df = pd.DataFrame(
            np.hstack((
                timestamp,
                fdata['cni'],
                fdata['U0'][:, np.newaxis],
                fdata['Cl3'],
                fdata['Zg3'],
                fdata['As3'][:, 0, :].squeeze(),
                fdata['As3'][:, 1, :].squeeze(),
                fdata['As3'][:, 2, :].squeeze(),
                fdata['As3'][:, 3, :].squeeze(),
                fdata['Is3'],
                fdata['Yl3'],
                fdata['Il3'],
            ))
        )
        df.columns = (
            'datetime',
            'a',
            'b',
            'U0',
            'Cl1',
            'Cl2',
            'Cl3',
            'Zg1',
            'Zg2',
            'Zg3',
            'ShuntVoltage1_1',
            'ShuntVoltage1_2',
            'ShuntVoltage1_3',
            'ShuntVoltage2_1',
            'ShuntVoltage2_2',
            'ShuntVoltage2_3',
            'ShuntVoltage3_1',
            'ShuntVoltage3_2',
            'ShuntVoltage3_3',
            'ShuntVoltage4_1',
            'ShuntVoltage4_2',
            'ShuntVoltage4_3',
            'Is1',
            'Is2',
            'Is3',
            'Yl1',
            'Yl2',
            'Yl3',
            'Il1',
            'Il2',
            'Il3',
        )

        df['datetime'] = pd.to_datetime(df['datetime'])
        # df['a'] = df['a'].astype(int)
        # df['b'] = df['b'].astype(int)
        # I THINK we can just sort a, b - the actual data should already
        # conform to the notation that the lower electrode number is
        # assigned to a
        df[['a', 'b']] = np.sort(df[['a', 'b']].values, axis=1)
        df['Cl1'] = df['Cl1'].astype(complex)
        df['Cl2'] = df['Cl2'].astype(complex)
        df['Cl3'] = df['Cl3'].astype(complex)
        df['Zg1'] = df['Zg1'].astype(complex)
        df['Zg2'] = df['Zg2'].astype(complex)
        df['Zg3'] = df['Zg3'].astype(complex)

        df['Yl1'] = df['Yl1'].astype(complex)
        df['Yl2'] = df['Yl2'].astype(complex)
        df['Yl3'] = df['Yl3'].astype(complex)

        for key in ('Il1', 'Il2', 'Il3'):
            df[key] = df[key].astype(complex)

        df['ShuntVoltage1_1'] = df['ShuntVoltage1_1'].astype(complex)
        df['ShuntVoltage1_2'] = df['ShuntVoltage1_2'].astype(complex)
        df['ShuntVoltage1_3'] = df['ShuntVoltage1_3'].astype(complex)

        df['ShuntVoltage2_1'] = df['ShuntVoltage2_1'].astype(complex)
        df['ShuntVoltage2_2'] = df['ShuntVoltage2_2'].astype(complex)
        df['ShuntVoltage2_3'] = df['ShuntVoltage2_3'].astype(complex)

        df['ShuntVoltage3_1'] = df['ShuntVoltage3_1'].astype(complex)
        df['ShuntVoltage3_2'] = df['ShuntVoltage3_2'].astype(complex)
        df['ShuntVoltage3_3'] = df['ShuntVoltage3_3'].astype(complex)

        df['ShuntVoltage4_1'] = df['ShuntVoltage4_1'].astype(complex)
        df['ShuntVoltage4_2'] = df['ShuntVoltage4_2'].astype(complex)
        df['ShuntVoltage4_3'] = df['ShuntVoltage4_3'].astype(complex)

        df['Is1'] = df['Is1'].astype(complex)
        df['Is2'] = df['Is2'].astype(complex)
        df['Is3'] = df['Is3'].astype(complex)

        df['Is'] = np.mean(df[['Is1', 'Is2', 'Is3']].values, axis=1)
        # "standard" injected current, in [mA]
        df['Iab'] = np.abs(df['Is']) * 1e3
        df['Iab'] = df['Iab'].astype(float)

        df['Il'] = np.mean(df[['Il1', 'Il2', 'Il3']].values, axis=1)
        # take absolute value and convert to mA
        df['Ileakage'] = np.abs(df['Il']) * 1e3
        df['Ileakage'] = df['Ileakage'].astype(float)

        df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']], axis=1)

        df['frequency'] = np.ones(df.shape[0]) * fdata['fm']
        dfl.append(df)

    df = pd.concat(dfl)

    return df


def _extract_emd(mat, **kwargs):
    """Extract the data from the EMD substruct, given a medusa-created MNU0-mat
    file

    Parameters
    ----------

    mat: matlab-imported struct

    """
    emd = mat['EMD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, emd.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = emd[f_id]
        # some consistency checks
        if len(fdata['nu']) == 2 and fdata['nu'].shape[1] == 2:
            raise Exception('Need MNU0 file, not a quadpole .mat file:')

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in fdata['Time'].squeeze()]
        ).T
        df = pd.DataFrame(
            np.hstack((
                timestamp,
                fdata['ni'],
                fdata['nu'][:, np.newaxis],
                fdata['Zt3'],
                fdata['Is3'],
                fdata['Il3'],
                fdata['Zg3'],
                fdata['As3'][:, 0, :].squeeze(),
                fdata['As3'][:, 1, :].squeeze(),
                fdata['As3'][:, 2, :].squeeze(),
                fdata['As3'][:, 3, :].squeeze(),
                fdata['Yg13'],
                fdata['Yg23'],
            )),
        )
        df.columns = (
            'datetime',
            'a',
            'b',
            'p',
            'Z1',
            'Z2',
            'Z3',
            'Is1',
            'Is2',
            'Is3',
            'Il1',
            'Il2',
            'Il3',
            'Zg1',
            'Zg2',
            'Zg3',
            'ShuntVoltage1_1',
            'ShuntVoltage1_2',
            'ShuntVoltage1_3',
            'ShuntVoltage2_1',
            'ShuntVoltage2_2',
            'ShuntVoltage2_3',
            'ShuntVoltage3_1',
            'ShuntVoltage3_2',
            'ShuntVoltage3_3',
            'ShuntVoltage4_1',
            'ShuntVoltage4_2',
            'ShuntVoltage4_3',
            'Yg13_1',
            'Yg13_2',
            'Yg13_3',
            'Yg23_1',
            'Yg23_2',
            'Yg23_3',
        )

        df['frequency'] = np.ones(df.shape[0]) * fdata['fm']

        # cast to correct type
        df['datetime'] = pd.to_datetime(df['datetime'])
        df['a'] = df['a'].astype(int)
        df['b'] = df['b'].astype(int)
        df['p'] = df['p'].astype(int)

        df['Z1'] = df['Z1'].astype(complex)
        df['Z2'] = df['Z2'].astype(complex)
        df['Z3'] = df['Z3'].astype(complex)

        df['Zg1'] = df['Zg1'].astype(complex)
        df['Zg2'] = df['Zg2'].astype(complex)
        df['Zg3'] = df['Zg3'].astype(complex)

        df['Is1'] = df['Is1'].astype(complex)
        df['Is2'] = df['Is2'].astype(complex)
        df['Is3'] = df['Is3'].astype(complex)

        df['Il1'] = df['Il1'].astype(complex)
        df['Il2'] = df['Il2'].astype(complex)
        df['Il3'] = df['Il3'].astype(complex)

        df['ShuntVoltage1_1'] = df['ShuntVoltage1_1'].astype(complex)
        df['ShuntVoltage1_2'] = df['ShuntVoltage1_2'].astype(complex)
        df['ShuntVoltage1_3'] = df['ShuntVoltage1_3'].astype(complex)

        df['ShuntVoltage2_1'] = df['ShuntVoltage2_1'].astype(complex)
        df['ShuntVoltage2_2'] = df['ShuntVoltage2_2'].astype(complex)
        df['ShuntVoltage2_3'] = df['ShuntVoltage2_3'].astype(complex)

        df['ShuntVoltage3_1'] = df['ShuntVoltage3_1'].astype(complex)
        df['ShuntVoltage3_2'] = df['ShuntVoltage3_2'].astype(complex)
        df['ShuntVoltage3_3'] = df['ShuntVoltage3_3'].astype(complex)

        df['ShuntVoltage4_1'] = df['ShuntVoltage4_1'].astype(complex)
        df['ShuntVoltage4_2'] = df['ShuntVoltage4_2'].astype(complex)
        df['ShuntVoltage4_3'] = df['ShuntVoltage4_3'].astype(complex)

        dfl.append(df)

    if len(dfl) == 0:
        return None
    df = pd.concat(dfl)

    # average swapped current injections here!
    # TODO

    # sort current injections
    condition = df['a'] > df['b']
    df.loc[condition, ['a', 'b']] = df.loc[condition, ['b', 'a']].values
    # for some reason we lose the integer casting of a and b here
    df['a'] = df['a'].astype(int)
    df['b'] = df['b'].astype(int)
    # change sign because we changed A and B
    df.loc[condition, ['Z1', 'Z2', 'Z3']] *= -1

    # average of Z1-Z3
    df['Zt'] = np.mean(df[['Z1', 'Z2', 'Z3']].values, axis=1)
    # we need to keep the sign of the real part
    sign_re = np.real(df['Zt']) / np.abs(np.real(df['Zt']))
    df['r'] = np.abs(df['Zt']) * sign_re
    # df['Zt_std'] = np.std(df[['Z1', 'Z2', 'Z3']].values, axis=1)

    df['Is'] = np.mean(df[['Is1', 'Is2', 'Is3']].values, axis=1)
    df['Il'] = np.mean(df[['Il1', 'Il2', 'Il3']].values, axis=1)
    df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']].values, axis=1)

    # "standard" injected current, in [mA]
    df['Iab'] = np.abs(df['Is']) * 1e3
    df['Iab'] = df['Iab'].astype(float)
    # df['Is_std'] = np.std(df[['Is1', 'Is2', 'Is3']].values, axis=1)

    # take absolute value and convert to mA
    df['Ileakage'] = np.abs(df['Il']) * 1e3
    df['Ileakage'] = df['Ileakage'].astype(float)

    return df



# coding=utf-8
"""
Testing utilities
"""

# import required for some doctests
from .containers import ERTContainer

# import os
import sys
from os.path import join, realpath

import matplotlib.pyplot as plt
ERTContainer


def test(target=None, show=False, onlydoctests=False, abort=False,
         verbose=True):
    """Run docstring examples and additional tests.

    Parameters
    ----------
    target : function or string, optional
        Function or method to test. By default everything is tested.
    show : boolean, optional
        Show matplotlib windows during test run. They will be closed
        automatically.
    onlydoctests : boolean, optional
        Run test files in ../tests as well.
    abort : boolean, optional
        Return correct exit code, e.g. abort documentation build when a test
        fails.
    """

    old_backend = plt.get_backend()
    if not show:
        plt.switch_backend("Agg")

    if target:
        if isinstance(target, str):
            # If target is a string, the code below will overwrite target
            # with the corresponding imported function, so that doctest works.
            import importlib
            mod_name, func_name = target.rsplit('.', 1)
            mod = importlib.import_module(mod_name)
            target = getattr(mod, func_name)

        import doctest
        doctest.run_docstring_examples(
            target, globals(),
            verbose=verbose,
            optionflags=doctest.ELLIPSIS,
            name=target.__name__
        )
        return

    try:
        import pytest
    except ImportError:
        raise ImportError(
            "pytest is required to run test suite. " "Try 'pip install" +
            "pytest'."
        )

    cwd = join(realpath(__path__[0]), '..')

    excluded = [
            # path...
    ]

    if onlydoctests:
        excluded.append("testing")

    cmd = ([
        "-v", "-rsxX", "--color", "yes", "--doctest-modules",
        "--durations", "5", cwd
    ])
    for directory in excluded:
        cmd.extend(["--ignore", join(cwd, directory)])

    exitcode = pytest.main(cmd)
    plt.switch_backend(old_backend)
    plt.close('all')
    if abort:
        sys.exit(exitcode)



import numpy as np
import pandas as pd

import reda.utils.mpl

plt, mpl = reda.utils.mpl.setup()


def plot_quadpole_evolution(dataobj, quadpole, cols, threshold=5,
                            rolling=False, ax=None):
    """Visualize time-lapse evolution of a single quadropole.

    Parameters
    ----------
    dataobj : :py:class:`pandas.DataFrame`
        DataFrame containing the data. Please refer to the documentation for
        required columns.
    quadpole : list of integers
        Electrode numbers of the the quadropole.
    cols : str
        The column/parameter to plot over time.
    threshold : float
        Allowed percentage deviation from the rolling standard deviation.
    rolling : bool
        Calculate rolling median values (the default is False).
    ax : mpl.axes
        Optional axes object to plot to.
    """
    if isinstance(dataobj, pd.DataFrame):
        df = dataobj
    else:
        df = dataobj.data

    subquery = df.query(
        'a == {0} and b == {1} and m == {2} and n == {3}'.format(*quadpole))
    # rhoa = subquery['rho_a'].values
    # rhoa[30] = 300
    # subquery['rho_a'] = rhoa

    if ax is not None:
        fig = ax.get_figure()
    else:
        fig, ax = plt.subplots(1, 1, figsize=(20 / 2.54, 7 / 2.54))

    ax.plot(
        subquery['timestep'],
        subquery[cols],
        '.',
        color='blue',
        label='valid data',
    )
    if rolling:
        # rolling mean
        rolling_m = subquery.rolling(3, center=True, min_periods=1).median()

        ax.plot(
            rolling_m['timestep'].values,
            rolling_m['rho_a'].values,
            '-',
            label='rolling median',
        )

        ax.fill_between(
            rolling_m['timestep'].values,
            rolling_m['rho_a'].values * (1 - threshold),
            rolling_m['rho_a'].values * (1 + threshold),
            alpha=0.4,
            color='blue',
            label='{0}\% confidence region'.format(threshold * 100),
        )

        # find all values that deviate by more than X percent from the
        # rolling_m
        bad_values = (np.abs(
            np.abs(subquery['rho_a'].values - rolling_m['rho_a'].values) /
            rolling_m['rho_a'].values) > threshold)

        bad = subquery.loc[bad_values]
        ax.plot(
            bad['timestep'].values,
            bad['rho_a'].values,
            '.',
            # s=15,
            color='r',
            label='discarded data',
        )

    ax.legend(loc='upper center', fontsize=6)
    # ax.set_xlim(10, 20)

    ax.set_ylabel(r'$\rho_a$ [$\Omega$m]')
    ax.set_xlabel('timestep')
    return fig, ax



import numpy as np
import pandas as pd

import pytest
from reda.utils.electrode_manager import electrode_manager


def test_user_supplied_sorter():

    def reversed_sorter(data):
        coordinates_sorted = data.sort_values(
            ['z', 'y', 'x'], ascending=False)
        coordinates_sorted.index = range(1, coordinates_sorted.shape[0] + 1)
        print('SORTED')
        print(coordinates_sorted)
        return coordinates_sorted

    # provide only positions of electrodes
    test_3d_positions = pd.DataFrame(
        (
            (2, 0, 0),
            (0, 0, 0),
            (1, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_3d_positions)

    # default behavior
    assert np.all(elecs.get_position_of_number(1) == [2, 0, 0])
    assert np.all(elecs.get_position_of_number(2) == [0, 0, 0])
    assert np.all(elecs.get_position_of_number(3) == [1, 0, 0])

    elecs.register_custom_ordering_scheme(reversed_sorter)

    assert np.all(elecs.get_position_of_number(1) == [2, 0, 0])
    assert np.all(elecs.get_position_of_number(2) == [1, 0, 0])
    assert np.all(elecs.get_position_of_number(3) == [0, 0, 0])


def test_get_electrode_position():
    # provide only positions of electrodes
    test_3d_positions = pd.DataFrame(
        (
            (2, 0, 0),
            (0, 0, 0),
            (1, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_3d_positions)

    with pytest.raises(AssertionError):
        elecs.get_position_of_number(0)

    assert np.all(elecs.get_position_of_number(1) == [2, 0, 0])
    assert np.all(elecs.get_position_of_number(2) == [0, 0, 0])
    assert np.all(elecs.get_position_of_number(3) == [1, 0, 0])

    with pytest.raises(AssertionError):
        elecs.get_position_of_number(4)

    with pytest.raises(AssertionError):
        elecs.get_position_of_number(10)

    assert elecs.get_electrode_numbers_for_positions([-10, 2, 4]) is None
    # 1D/2D/3D
    assert elecs.get_electrode_numbers_for_positions([2]) == 1
    assert elecs.get_electrode_numbers_for_positions([2, 0]) == 1
    assert elecs.get_electrode_numbers_for_positions([2, 0, 0]) == 1

    assert elecs.get_electrode_numbers_for_positions([0, 0, 0]) == 2
    assert elecs.get_electrode_numbers_for_positions([1, 0, 0]) == 3

    # multiple 1D
    assert np.all(
        elecs.get_electrode_numbers_for_positions([[2, ], [0, ]]) == [1, 2]
    )


def test_get_electrode_numbers_for_positions():
    # provide only positions of electrodes
    test_3d_positions = pd.DataFrame(
        (
            (2, 0, 0),
            (0, 0, 0),
            (1, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_3d_positions)

    # test as is
    elecs.set_ordering_to_as_is()
    assert elecs.get_electrode_numbers_for_positions([2, 0, 0]) == 0
    assert elecs.get_electrode_numbers_for_positions([0, 0, 0]) == 1
    assert elecs.get_electrode_numbers_for_positions([1, 0, 0]) == 2

    assert elecs.get_electrode_numbers_for_positions([2, 0]) == 0, '2D, pos 1'
    assert elecs.get_electrode_numbers_for_positions([0, 0]) == 1, '2D, pos 2'
    assert elecs.get_electrode_numbers_for_positions([1, 0]) == 2, '2D, pos 3'

    assert np.all(
        elecs.get_electrode_numbers_for_positions(
            [[2, 0], [0, 0]]
        ) == [0, 1]
    ), 'multiple positions, nested list'


def test_sorters():
    # provide only positions of electrodes
    test_3d_positions = pd.DataFrame(
        (
            (2, 0, 0),
            (0, 0, 0),
            (1, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_3d_positions)

    # test as is
    elecs.set_ordering_to_as_is()
    assert elecs.get_electrode_numbers_for_positions([2, 0, 0]) == 0
    assert elecs.get_electrode_numbers_for_positions([0, 0, 0]) == 1
    assert elecs.get_electrode_numbers_for_positions([1, 0, 0]) == 2

    assert elecs.get_electrode_numbers_for_positions([2, 0]) == 0
    assert elecs.get_electrode_numbers_for_positions([0, 0]) == 1
    assert elecs.get_electrode_numbers_for_positions([1, 0]) == 2

    # test as is plus one
    elecs.set_ordering_to_as_is_plus_one()
    assert elecs.get_electrode_numbers_for_positions([2, 0, 0]) == 1
    assert elecs.get_electrode_numbers_for_positions([0, 0, 0]) == 2
    assert elecs.get_electrode_numbers_for_positions([1, 0, 0]) == 3

    assert elecs.get_electrode_numbers_for_positions([2, 0]) == 1
    assert elecs.get_electrode_numbers_for_positions([0, 0]) == 2
    assert elecs.get_electrode_numbers_for_positions([1, 0]) == 3

    # test sorted
    elecs.set_ordering_to_sort_zyx()
    assert elecs.get_electrode_numbers_for_positions([2, 0, 0]) == 3
    assert elecs.get_electrode_numbers_for_positions([0, 0, 0]) == 1
    assert elecs.get_electrode_numbers_for_positions([1, 0, 0]) == 2

    assert elecs.get_electrode_numbers_for_positions([2, 0]) == 3
    assert elecs.get_electrode_numbers_for_positions([0, 0]) == 1
    assert elecs.get_electrode_numbers_for_positions([1, 0]) == 2


def test_duplicates():
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_3d_positions)
    # add a second time
    elecs.add_by_position(test_3d_positions)
    elecs.add_by_position([0, 0, 0])
    # import IPython
    # IPython.embed()

    assert elecs.electrode_positions.shape == test_3d_positions.shape


def test_add_positions():
    test_pandas_3d = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_pandas_3d)
    assert np.all(test_pandas_3d.values == elecs.electrode_positions.values)

    test_pandas_2d = pd.DataFrame(
        (
            (0, 0),
            (1, 0),
            (2, 0),
        ),
        columns=['x', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_pandas_2d)
    assert np.all(test_pandas_3d.values == elecs.electrode_positions.values)

    test_pandas_1d = pd.DataFrame(
        (
            (0),
            (1),
            (2),
        ),
        columns=['x'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_pandas_1d)
    assert np.all(test_pandas_3d.values == elecs.electrode_positions.values)

    # lists
    elecs = electrode_manager()

    elecs.add_by_position([[1, ], [2, ]])
    elecs.add_by_position([4.5, 4.7])
    elecs.add_by_position([[5.5, 0], [7.0, 0]])
    elecs.add_by_position([[8, 0, 0], ])
    elecs.add_by_position([[9, 0, 0], [10, 0, 0]])

    result_expected = np.array((
        (1, 0, 0),
        (2, 0, 0),
        (4.5, 0, 0),
        (4.7, 0, 0),
        (5.5, 0, 0),
        (7.0, 0, 0),
        (8.0, 0, 0),
        (9.0, 0, 0),
        (10.0, 0, 0),
    ))
    assert np.all(result_expected == elecs.electrode_positions.values)

    # numpy arrays
    elecs = electrode_manager()
    # 1D
    elecs.add_by_position(np.array([[1.5, 1.7, 1.9, 1.95], ]).T)

    elecs.add_by_position(np.array([3.5, 0])[np.newaxis, :])
    elecs.add_by_position(np.array([[4.5, 0], [7.0, 0]]))
    elecs.add_by_position(np.array([8, 0, 0])[np.newaxis, :])
    elecs.add_by_position(np.array([[9, 0, 0], [10, 0, 0]]))

    result_expected = np.array((
        (1.5, 0, 0),
        (1.7, 0, 0),
        (1.9, 0, 0),
        (1.95, 0, 0),
        (3.5, 0, 0),
        (4.5, 0, 0),
        (7.0, 0, 0),
        (8.0, 0, 0),
        (9.0, 0, 0),
        (10.0, 0, 0),
    ))
    assert np.all(result_expected == elecs.electrode_positions.values)


def test_conform_to_regular_x_spacing():
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (2, 0, 0),
            (4, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    result_expected = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
            (3, 0, 0),
            (4, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    # this is important, we want a sorted output
    elecs.set_ordering_to_sort_zyx()
    elecs.add_by_position(test_3d_positions)
    elecs.conform_to_regular_x_spacing(spacing_x=1)

    assert np.all(result_expected.values == elecs.electrode_positions.values)

    # new test
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (2, 0, 0),
            (4, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    result_expected = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
            (3, 0, 0),
            (4, 0, 0),
            (5, 0, 0),
            (6, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    # this is important, we want a sorted output
    elecs.set_ordering_to_sort_zyx()
    elecs.add_by_position(test_3d_positions)
    elecs.conform_to_regular_x_spacing(spacing_x=1, nr_electrodes=7)
    assert np.all(result_expected.values == elecs.electrode_positions.values)


def test_conform_to_regular_x_spacing_exception():
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (2.5, 0, 0),
            (4, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_3d_positions)
    with pytest.raises(Exception):
        elecs.conform_to_regular_x_spacing(spacing_x=1)


def test_reflect_on_x():
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_3d_positions)
    elecs.reflect_on_position_x(10)
    result_expected = pd.DataFrame(
        (
            (10 - 0, 0, 0),
            (10 - 1, 0, 0),
            (10 - 2, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    assert np.all(result_expected.values == elecs.electrode_positions.values)


def test_electrode_numbers():
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(test_3d_positions)
    number_assignments = elecs.get_all_electrode_numbers()
    assert np.all(number_assignments.index.values.tolist() == [1, 2, 3])


def test_bad_input():
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
        ),
        columns=['a', 'y', 'z'],
    )
    elecs = electrode_manager()
    # we expect this function to raise an error
    try:
        elecs.add_by_position(test_3d_positions)
    except AssertionError:
        pass


def test_simple_ordering_3d():
    # provide only positions of electrodes
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
            (3, 0, 0),
            (4, 0, 0),
            (5, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    ).astype(float)

    elecs = electrode_manager()
    elecs.set_ordering_to_sort_zyx()
    elecs.add_by_position(test_3d_positions)

    for seed in range(50):
        rng = np.random.default_rng(seed)
        c = elecs.electrode_positions.copy()
        new_index = c.index.values
        rng.shuffle(new_index)
        c_unsorted = c.set_index(new_index).sort_index()

        c_sorted = elecs.sorter(c_unsorted)
        assert np.all(test_3d_positions.values == c_sorted.values)


def test_simple_ordering_2d():
    # provide only positions of electrodes
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
            (3, 0, 0),
            (4, 0, 0),
            (5, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    test_2d_positions = test_3d_positions[['x', 'z']]

    elecs = electrode_manager()
    elecs.set_ordering_to_sort_zyx()

    elecs.add_by_position(test_2d_positions)

    for seed in range(50):
        rng = np.random.default_rng(seed)
        c = elecs.electrode_positions.copy()
        new_index = c.index.values
        rng.shuffle(new_index)
        c_unsorted = c.set_index(new_index).sort_index()

        c_sorted = elecs.sorter(c_unsorted)
        assert np.all(test_3d_positions.values == c_sorted.values)


def test_number_assignment():
    # provide only positions of electrodes
    test_3d_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
            (3, 0, 0),
            (4, 0, 0),
            (5, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.set_ordering_to_as_is_plus_one()
    elecs.add_by_position(test_3d_positions)

    input_coords = pd.DataFrame(
        (
            (5, 0, 0),
            (0, 0, 0),
            (3, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    expected_numbers = [6, 1, 4]
    elec_numbers = elecs.get_electrode_numbers_for_positions(input_coords)
    assert expected_numbers == elec_numbers.tolist()


def test_fixed_assignments():
    fixed_assignments = pd.DataFrame(
        (
            (1, 0, 0, 0),
            (2, 1, 0, 0),
            (10, 2, 0, 0),
            (11, 3, 0, 0),
            (7, 4, 0, 0),
            (8, 5, 0, 0),
        ),
        columns=['electrode_number', 'x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_fixed_assignments(fixed_assignments)

    assert elecs.get_electrode_numbers_for_positions([2, 0, 0]) == 10
    assert elecs.get_electrode_numbers_for_positions([4, 0, 0]) == 7
    assert np.all(
        elecs.get_position_of_number(10).values == np.array((2, 0, 0))
    )

    fixed2 = fixed_assignments.set_index('electrode_number')
    elecs = electrode_manager()
    elecs.add_fixed_assignments(fixed2)

    assert elecs.get_electrode_numbers_for_positions([2, 0, 0]) == 10
    assert elecs.get_electrode_numbers_for_positions([4, 0, 0]) == 7
    assert np.all(
        elecs.get_position_of_number(10).values == np.array((2, 0, 0))
    )


def test_replace_coordinates_with_fixed_regular_grid_x():
    electrode_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
            (3, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)

    elecs.replace_coordinates_with_fixed_regular_grid_x(5)

    expected_result = pd.DataFrame(
        (
            (0, 0, 0),
            (5, 0, 0),
            (10, 0, 0),
            (15, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    assert np.all(
        expected_result.values == elecs.electrode_positions.values
    ), 'x position replacement did not work'


def test_replace_one_coordinate():
    electrode_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
            (3, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)

    with pytest.raises(AssertionError):
        elecs.replace_coordinate_of_electrode_number(5, [10, 0, 0])

    elecs.replace_coordinate_of_electrode_number(4, [10, 0, 0])
    assert np.all(elecs.get_position_of_number(4) == [10, 0, 0])

    elecs.replace_coordinate_of_electrode_number(1, [555, 0])
    assert np.all(elecs.get_position_of_number(1) == [555, 0, 0])


def test_callable_class():
    electrode_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
            (3, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )
    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)

    assert np.all(
        electrode_positions.values == elecs().values
    )


def test_merging():
    positions1 = pd.DataFrame(
        (
            (1, 0, 0, 0),
            (2, 1, 0, 0),
            (3, 2, 0, 0),
            # note gap here in positions!
            (4, 4, 0, 0),
        ),
        columns=['electrode_number', 'x', 'y', 'z'],
    )
    data1 = pd.DataFrame(
        (
            (1, 2, 4, 3),
            (1, 4, 3, 2),

        ),
        columns=['a', 'b', 'm', 'n'],
    )

    positions2 = pd.DataFrame(
        (
            (1, 3, 0, 0),
            (2, 4, 0, 0),
            (3, 5, 0, 0),
            (4, 6, 0, 0),
        ),
        columns=['electrode_number', 'x', 'y', 'z'],
    )
    data2 = pd.DataFrame(
        (
            (1, 2, 4, 3),
            (1, 4, 3, 2),

        ),
        columns=['a', 'b', 'm', 'n'],
    )

    elecs = electrode_manager()

    positions_aligned, data1_aligned, data2_aligned = elecs.align_assignments(
        positions1,
        positions2,
        data1,
        data2,
    )

    positions_expected = pd.DataFrame(
        (
            (1, 0, 0, 0),
            (2, 1, 0, 0),
            (3, 2, 0, 0),
            (4, 4, 0, 0),
            (5, 3, 0, 0),
            (6, 5, 0, 0),
            (7, 6, 0, 0),
        ),
        columns=['electrode_number', 'x', 'y', 'z'],
    ).astype(float).set_index('electrode_number')

    assert np.all(
        positions_expected == positions_aligned
    ), 'aligned positions'

    data2_expected = pd.DataFrame(
        (
            (5, 4, 7, 6),
            (5, 7, 6, 4),

        ),
        columns=['a', 'b', 'm', 'n'],
    )

    assert np.all(data1_aligned == data1)
    assert np.all(data2_aligned == data2_expected)


def test_merging2():
    pos1 = pd.DataFrame(np.arange(0, 2, 0.2), columns=['x', ])
    pos1['y'] = 0
    pos1['z'] = 0
    pos1.index.name = 'electrode_number'

    pos2 = pd.DataFrame(np.arange(1, 3, 0.2), columns=['x', ])
    pos2['y'] = 0
    pos2['z'] = 0
    pos2.index.name = 'electrode_number'

    data1 = pd.DataFrame(
        (
            (1, 2, 4, 3),
            (1, 4, 3, 2),

        ),
        columns=['a', 'b', 'm', 'n'],
    )

    elecs = electrode_manager()

    positions_aligned, data1_aligned, data2_aligned = elecs.align_assignments(
        pos1,
        pos2,
        data1,
        data1.copy(),
    )

    pos_expected = pd.DataFrame(np.arange(0, 3, 0.2), columns=['x', ])
    pos_expected['y'] = 0
    pos_expected['z'] = 0

    assert np.all(
        np.isclose(pos_expected.values, positions_aligned.values)
    ), 'aligned positions'

    # data2_expected = pd.DataFrame(
    #     (
    #         (5, 4, 7, 6),
    #         (5, 7, 6, 4),

    #     ),
    #     columns=['a', 'b', 'm', 'n'],
    # )

    # assert np.all(data1_aligned == data1)
    # assert np.all(data2_aligned == data2_expected)


def test_shift_positions_by_xyz():
    electrode_positions = pd.DataFrame(
        (
            (0, 0, 0),
            (1, 0, 0),
            (2, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )

    result_expected = pd.DataFrame(
        (
            (10, 0, 0),
            (11, 0, 0),
            (12, 0, 0),
        ),
        columns=['x', 'y', 'z'],
    )

    # list
    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)
    elecs.shift_positions_xyz([10, ])
    assert np.all(elecs().values == result_expected)

    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)
    elecs.shift_positions_xyz([10, 0])
    assert np.all(elecs().values == result_expected)

    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)
    elecs.shift_positions_xyz([10, 0, 0])
    assert np.all(elecs().values == result_expected)

    # array
    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)
    elecs.shift_positions_xyz(np.array([10, ]))
    assert np.all(elecs().values == result_expected)

    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)
    elecs.shift_positions_xyz(np.array([10, 0]))
    assert np.all(elecs().values == result_expected)

    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)
    elecs.shift_positions_xyz(np.array([10, 0, 0]))
    assert np.all(elecs().values == result_expected)

    # assertions
    elecs = electrode_manager()
    elecs.add_by_position(electrode_positions)
    with pytest.raises(Exception):
        elecs.shift_positions_xyz(np.array([10, 0, 0, 0]))
    with pytest.raises(Exception):
        elecs.shift_positions_xyz(np.array([[10, 0, 0], [2, 0, 0]]))


if __name__ == '__main__':
    # test_merging()
    test_merging2()
    # test_shift_positions_by_xyz()
    pass



"""Container tests for the ERT container

# Missing test cases

* Add data with and without electrode positions
* Add data with and without timesteps (i.e., import multiple data for one time
  step, or for multiple ones)
* How to proceed if electrode positions were already imported, but a newly
  added dataset does not provide electrode coordinates?
  -> throw an error if there are any electrode numbers to accounted for yet
  -> make sure the documentation knows how to add electrode numbers
* test the verbose switch. Is this the same as debugging?
* How to test the journal?
* Check the initialization with a dataframe (or better: try to initialize with
  something else and check for the exception)


"""
import pandas as pd

import pytest

import reda


def test_init():
    """test initializing an empty ERT container"""
    container = reda.ERT()
    assert isinstance(container, reda.ERT)


def test_bad_init():
    # anything but a DataFrame
    bad_input_data = 1
    with pytest.raises(Exception):
        reda.ERT(data=bad_input_data)


def test_init_with_data():
    """test initializing an ERT container and provide good data"""
    df = pd.DataFrame(
        [
            # normals
            (0, 1, 2, 4, 3, 1.1),
            (0, 1, 2, 5, 4, 1.2),
            (0, 1, 2, 6, 5, 1.3),
            (0, 1, 2, 7, 6, 1.4),
            (0, 2, 3, 5, 4, 1.5),
            (0, 2, 3, 6, 5, 1.6),
            (0, 2, 3, 7, 6, 1.7),
            (0, 3, 4, 6, 5, 1.8),
            (0, 3, 4, 7, 6, 1.9),
            (0, 4, 5, 7, 6, 2.0),
        ],
        columns=['timestep', 'a', 'b', 'm', 'n', 'r'],
    )
    container_good = reda.ERT(data=df)
    assert container_good.data.shape[0] == df.shape[0]



# -*- coding: utf-8 -*-
"""Importer functions for data files measured with the SIP-04 SIP system,
developed at the research center Jülich.

The result files contain a lot of different, sometimes redundant parameters.

========= ====================================================
Parameter Explanation/Meaning of Parameters:
========= ====================================================
..._1     First measurement
..._2     Second measurement
..._3     Third measurement
..._m     Mean value of that three measurements
..._std   Standard deviation of three three measurements
Zm        complex value of transfer impedance
Zm_mAbs   Absolute value of Zm_m
Zm_mPhi   Phase shift of Zm_m
Zm_mRe    Real part of Zm_m
Zm_mIm    Imaginary Part of Zm_m
Zm_AbsStd Standard deviation of all 3 Zm_Abs values
Zm_PhiStd Standard deviation of all 3 Zm_Phi values
========= ====================================================
"""
import scipy.io
import pandas as pd
import numpy as np
import os


def import_sip04_data(data_filename):
    """Import RELEVANT data from the result files. Refer to the function
    :func:`reda.importers.sip04.import_sip04_data_all` for an importer that
    imports ALL data.

    Exported parameters:

    ================== ========================================================
    key                description
    ================== ========================================================
    a                  First current electrode
    b                  Second current electrode
    m                  First potential electrode
    n                  Second potential electrode
    frequency          Measurement frequency
    Temp_1             Temperature sensor 1 (optional)
    Temp_2             Temperature sensor 2 (optional)
    zt                 Complex Transfer Impedance (the measurement), mean value
    r                  Magnitude of mean measurements (=|zt|)
    rpha               Resistance phase [mrad]
    zt_1               Complex Transfer Impedance, first repetition
    zt_2               Complex Transfer Impedance, second repetition
    zt_3               Complex Transfer Impedance, third repetition
    ContactResistance  Contact resistance (mean value)
    ShuntResistance    Shunt resistance used [Ohm]
    ================== ========================================================

    Parameters
    ----------
    data_filename : string
        Path to .mat or .csv file containing SIP-04 measurement results. Note
        that the .csv file does not contain all data contained in the .mat
        file!

    Returns
    -------
    df : :py:class:`pandas.DataFrame`
        The data, contained in a DataFrame

    """
    df_all = import_sip04_data_all(data_filename)
    columns_to_keep = list(
        set(df_all.columns) &
        set([
            'a', 'b', 'm', 'n',
            'frequency',
            'Temp_1', 'Temp_2',
            'Zm_1', 'Zm_2', 'Zm_3',
            'Zg_m',
            'zt',
            'Rs',
            'r',
            'rpha',
        ])
    )
    # import IPython
    # IPython.embed()
    df = df_all[columns_to_keep]
    df = df.rename(columns={
        'Rs': 'ShuntResistance',
        'Zg_m': 'ContactResistance',
        'Zm_1': 'zt_1',
        'Zm_2': 'zt_2',
        'Zm_3': 'zt_3',
    })

    return df


def import_sip04_data_all(data_filename):
    """Import ALL data from the result files

    Parameters
    ----------
    data_filename : string
        Path to .mat or .csv file containing SIP-04 measurement results. Note
        that the .csv file does not contain all data contained in the .mat
        file!

    Returns
    -------
    df_all : :py:class:`pandas.DataFrame`
        The data, contained in a DataFrame
    """
    filename, fformat = os.path.splitext(data_filename)

    if fformat == '.csv':
        print('Import SIP04 data from .csv file')
        df_all = _import_csv_file(data_filename)
    elif fformat == '.mat':
        print('Import SIP04 data from .mat file')
        df_all = _import_mat_file(data_filename)
    else:
        print('Please use .csv or .mat format.')
        df_all = None

    return df_all


def _import_mat_file(mat_filename):
    """
    """
    mat = scipy.io.loadmat(mat_filename, squeeze_me=True)  # loading mat file

    # loading all parameters from the .mat file to the DataFrame 'df'
    df = pd.DataFrame(mat['fm'], columns=['fm'])  # frequencies
    # import IPython
    # IPython.embed()
    if mat['Temp'].size > 0:
        df['Temp_1'] = pd.Series(mat['Temp'][:, 0], index=df.index)
        df['Temp_2'] = pd.Series(mat['Temp'][:, 1], index=df.index)
        df['Temp_m'] = pd.Series(
            np.mean([df['Temp_1'], df['Temp_2']], axis=0),
            index=df.index
        )

    df['Time'] = pd.Series(mat['Time'], index=df.index)
    df['Zm_1'] = pd.Series(mat['Zm'][:, 0], index=df.index)
    df['Zm_2'] = pd.Series(mat['Zm'][:, 1], index=df.index)
    df['Zm_3'] = pd.Series(mat['Zm'][:, 2], index=df.index)
    df['Rs'] = pd.Series(
        np.array(mat['fm'].size * [mat['Rs']]),
        index=df.index
    )  # single value of RS for all enteries
    df['Zg_1'] = pd.Series(mat['Zg'][:, 0], index=df.index)
    df['Zg_2'] = pd.Series(mat['Zg'][:, 1], index=df.index)
    df['Zg_3'] = pd.Series(mat['Zg'][:, 2], index=df.index)
    df['Z12_1'] = pd.Series(mat['Z12'][:, 0], index=df.index)
    df['Z12_2'] = pd.Series(mat['Z12'][:, 1], index=df.index)
    df['Z12_3'] = pd.Series(mat['Z12'][:, 2], index=df.index)
    df['Z14_1'] = pd.Series(mat['Z14'][:, 0], index=df.index)
    df['Z14_2'] = pd.Series(mat['Z14'][:, 1], index=df.index)
    df['Z14_3'] = pd.Series(mat['Z14'][:, 2], index=df.index)
    df['Z34_1'] = pd.Series(mat['Z34'][:, 0], index=df.index)
    df['Z34_2'] = pd.Series(mat['Z34'][:, 1], index=df.index)
    df['Z34_3'] = pd.Series(mat['Z34'][:, 2], index=df.index)
    df['Z12_m'] = pd.Series(mat['Z12m'], index=df.index)
    df['Z14_m'] = pd.Series(mat['Z14m'], index=df.index)
    df['Z34_m'] = pd.Series(mat['Z34m'], index=df.index)
    df['Zg_m'] = pd.Series(mat['Zgm'], index=df.index)
    df['Zm_m'] = pd.Series(mat['Zmm'], index=df.index)
    df['Ug1_1'] = pd.Series(mat['Ug'][:, 0, 0], index=df.index)
    df['Ug1_2'] = pd.Series(mat['Ug'][:, 0, 1], index=df.index)
    df['Ug1_3'] = pd.Series(mat['Ug'][:, 0, 2], index=df.index)
    df['Ug2_1'] = pd.Series(mat['Ug'][:, 1, 0], index=df.index)
    df['Ug2_2'] = pd.Series(mat['Ug'][:, 1, 1], index=df.index)
    df['Ug2_3'] = pd.Series(mat['Ug'][:, 1, 2], index=df.index)
    df['Ug3_1'] = pd.Series(mat['Ug'][:, 2, 0], index=df.index)
    df['Ug3_2'] = pd.Series(mat['Ug'][:, 2, 1], index=df.index)
    df['Ug3_3'] = pd.Series(mat['Ug'][:, 2, 2], index=df.index)
    df['Ug4_1'] = pd.Series(mat['Ug'][:, 3, 0], index=df.index)
    df['Ug4_2'] = pd.Series(mat['Ug'][:, 3, 1], index=df.index)
    df['Ug4_3'] = pd.Series(mat['Ug'][:, 3, 2], index=df.index)
    df['Us1_1'] = pd.Series(mat['Us'][:, 0, 0], index=df.index)
    df['Us1_2'] = pd.Series(mat['Us'][:, 0, 1], index=df.index)
    df['Us1_3'] = pd.Series(mat['Us'][:, 0, 2], index=df.index)
    df['Us2_1'] = pd.Series(mat['Us'][:, 1, 0], index=df.index)
    df['Us2_2'] = pd.Series(mat['Us'][:, 1, 1], index=df.index)
    df['Us2_3'] = pd.Series(mat['Us'][:, 1, 2], index=df.index)
    df['Us3_1'] = pd.Series(mat['Us'][:, 2, 0], index=df.index)
    df['Us3_2'] = pd.Series(mat['Us'][:, 2, 1], index=df.index)
    df['Us3_3'] = pd.Series(mat['Us'][:, 2, 2], index=df.index)
    df['Us4_1'] = pd.Series(mat['Us'][:, 3, 0], index=df.index)
    df['Us4_2'] = pd.Series(mat['Us'][:, 3, 1], index=df.index)
    df['Us4_3'] = pd.Series(mat['Us'][:, 3, 2], index=df.index)
    df['Us1_m'] = pd.Series(mat['Usm'][:, 0], index=df.index)
    df['Us2_m'] = pd.Series(mat['Usm'][:, 1], index=df.index)
    df['Us3_m'] = pd.Series(mat['Usm'][:, 2], index=df.index)
    df['Us4_m'] = pd.Series(mat['Usm'][:, 3], index=df.index)

    # calculate other values, e.g. used in the .csv-file
    df['Zm_mAbs'] = pd.Series(np.abs(df['Zm_m']), index=df.index)
    Zm_1Abs = np.abs(df['Zm_1'])
    Zm_2Abs = np.abs(df['Zm_2'])
    Zm_3Abs = np.abs(df['Zm_3'])
    df['Zm_AbsStd'] = pd.Series(
        np.std([Zm_1Abs, Zm_2Abs, Zm_3Abs], axis=0),
        index=df.index
    )
    df['Zm_mPhi'] = pd.Series(np.angle(df['Zm_m']), index=df.index)
    Zm_1Phi = np.angle(df['Zm_1'])
    Zm_2Phi = np.angle(df['Zm_2'])
    Zm_3Phi = np.angle(df['Zm_3'])
    df['Zm_PhiStd'] = pd.Series(
        np.std([Zm_1Phi, Zm_2Phi, Zm_3Phi], axis=0),
        index=df.index
    )
    df['Zm_mRe'] = pd.Series(np.real(df['Zm_m']), index=df.index)
    df['Zm_mIm'] = pd.Series(np.imag(df['Zm_m']), index=df.index)
    df['Z12_mAbs'] = pd.Series(np.abs(df['Z12_m']), index=df.index)
    df['Z12_mPhi'] = pd.Series(np.angle(df['Z12_m']), index=df.index)
    df['Z34_mAbs'] = pd.Series(np.abs(df['Z34_m']), index=df.index)
    df['Z34_mPhi'] = pd.Series(np.angle(df['Z34_m']), index=df.index)
    df['Z14_mAbs'] = pd.Series(np.abs(df['Z14_m']), index=df.index)
    df['Z14_mPhi'] = pd.Series(np.angle(df['Z14_m']), index=df.index)
    df['Ug1_m'] = pd.Series(np.mean([df['Ug1_1'], df['Ug1_2'], df['Ug1_3']],
                                    axis=0),
                            index=df.index)
    df['Ug2_m'] = pd.Series(np.mean([df['Ug2_1'], df['Ug2_2'], df['Ug2_3']],
                                    axis=0),
                            index=df.index)
    df['Ug3_m'] = pd.Series(np.mean([df['Ug2_1'], df['Ug2_2'], df['Ug2_3']],
                                    axis=0),
                            index=df.index)
    df['Ug4_m'] = pd.Series(
        np.mean([df['Ug2_1'], df['Ug2_2'], df['Ug2_3']], axis=0),
        index=df.index
    )
    df['a'] = 1
    df['b'] = 4
    df['m'] = 2
    df['n'] = 3
    df['zt'] = df['Zm_m']
    # compute magnitude and phase [in mrad]
    df['r'] = np.abs(df['zt'])
    df['rpha'] = np.arctan2(np.imag(df['zt']), np.real(df['zt'])) * 1000
    df['frequency'] = df['fm']

    return df


def _import_csv_file(csv_filename):
    """
    """
    # first, getting all csv-data into a DataFrame
    # csv_filename = '08112013_Elbsandstein_00703A.csv'

    df = pd.read_csv(csv_filename,
                     sep=';',
                     skipinitialspace=True,
                     skip_blank_lines=False)

    newRow1 = int(df[df['f'] == 'f'].index[0])
    newRow2 = int(df[df['f'] == 'f'].index[1])
    newRow3 = int(df[df['f'] == 'f'].index[2])

    df1 = pd.read_csv(csv_filename,
                      sep=';',
                      skipinitialspace=True,
                      skiprows=0,
                      nrows=newRow1 - 1)

    df2 = pd.read_csv(csv_filename,
                      sep=';',
                      skipinitialspace=True,
                      skiprows=newRow1,
                      nrows=newRow2 - newRow1 - 2)

    df3 = pd.read_csv(csv_filename,
                      sep=';',
                      skipinitialspace=True,
                      skiprows=newRow2,
                      nrows=newRow3 - newRow2 - 2)

    df4 = pd.read_csv(csv_filename,
                      sep=';',
                      skipinitialspace=True,
                      skiprows=newRow3)

    df_merged = pd.concat([df1, df2, df3, df4], axis=1)
    df_merged = df_merged.T.drop_duplicates().T
    df_merged = df_merged.drop(['Unnamed: 6'], axis=1)
    df_merged = df_merged.rename(
        index=str,
        columns={
            'f': 'fm',
            'Abs(Zm)': 'Zm_mAbs',
            'Std(Abs)': 'Zm_AbsStd',
            'Phi(Zm)': 'Zm_mPhi',
            'Std(Phi)': 'Zm_PhiStd',
            'Re(Zm)': 'Zm_mRe',
            'Im(Zm)': 'Zm_mIm',
            'Time [s]': 'Time',
            'Abs(Z12)': 'Z12_mAbs',
            'Phi(Z12)': 'Z12_mPhi',
            'Abs(Z34)': 'Z34_mAbs',
            'Phi(Z34)': 'Z34_mPhi',
            'Abs(Z14)': 'Z14_mAbs',
            'Phi(Z14)': 'Z14_mPhi',
            'Ug1': 'Ug1_m',
            'Ug2': 'Ug2_m',
            'Ug3': 'Ug3_m',
            'Ug4': 'Ug4_m',
            'Us1': 'Us1_m',
            'Us2': 'Us2_m',
            'Us3': 'Us3_m',
            'Us4': 'Us4_m'
        }
    )

    # filling the final DataFrame:
    for column in ('Temp_1', 'Temp_2', 'Temp_m', 'Zm_1', 'Zm_2', 'Zm_3'):
        df_merged[column] = np.nan

    df_merged['Rs'] = np.nan
    df_merged['Zg_1'] = np.nan
    df_merged['Zg_2'] = np.nan
    df_merged['Zg_3'] = np.nan
    df_merged['Z12_1'] = np.nan
    df_merged['Z12_2'] = np.nan
    df_merged['Z12_3'] = np.nan
    df_merged['Z14_1'] = np.nan
    df_merged['Z14_2'] = np.nan
    df_merged['Z14_3'] = np.nan
    df_merged['Z34_1'] = np.nan
    df_merged['Z34_2'] = np.nan
    df_merged['Z34_3'] = np.nan
    df_merged['Ug1_1'] = np.nan
    df_merged['Ug1_2'] = np.nan
    df_merged['Ug1_3'] = np.nan
    df_merged['Ug2_1'] = np.nan
    df_merged['Ug2_2'] = np.nan
    df_merged['Ug2_3'] = np.nan
    df_merged['Ug3_1'] = np.nan
    df_merged['Ug3_2'] = np.nan
    df_merged['Ug3_3'] = np.nan
    df_merged['Ug4_1'] = np.nan
    df_merged['Ug4_2'] = np.nan
    df_merged['Ug4_3'] = np.nan
    df_merged['Us1_1'] = np.nan
    df_merged['Us1_2'] = np.nan
    df_merged['Us1_3'] = np.nan
    df_merged['Us2_1'] = np.nan
    df_merged['Us2_2'] = np.nan
    df_merged['Us2_3'] = np.nan
    df_merged['Us3_1'] = np.nan
    df_merged['Us3_2'] = np.nan
    df_merged['Us3_3'] = np.nan
    df_merged['Us4_1'] = np.nan
    df_merged['Us4_2'] = np.nan
    df_merged['Us4_3'] = np.nan
    df_merged['Zg_m'] = np.nan

    # calculating other values, e.g. used in the .mat-file
    df_merged['Z12_m'] = pd.Series(
        df_merged['Z12_mAbs'] * np.exp(1j * df_merged['Z12_mPhi']),
        index=df_merged.index
    )
    df_merged['Z14_m'] = pd.Series(
        df_merged['Z14_mAbs'] * np.exp(1j * df_merged['Z14_mPhi']),
        index=df_merged.index
    )
    df_merged['Z34_m'] = pd.Series(
        df_merged['Z34_mAbs'] * np.exp(1j * df_merged['Z34_mPhi']),
        index=df_merged.index
    )
    df_merged['Zm_m'] = pd.Series(
        df_merged['Zm_mAbs'] * np.exp(1j * df_merged['Zm_mPhi']),
        index=df_merged.index
    )

    df_merged['a'] = 1
    df_merged['b'] = 4
    df_merged['m'] = 2
    df_merged['n'] = 3

    df_merged['zt'] = df_merged['Zm_m']

    # compute magnitude and phase [in mrad]
    df_merged['r'] = np.abs(df_merged['zt'])
    df_merged['rpha'] = np.arctan2(
        np.imag(df_merged['zt']),
        np.real(df_merged['zt'])
    ) * 1000

    df_merged['frequency'] = df_merged['fm']

    return df_merged


"""
There are different results, depending on the complex calculation:
comp1 = df_merged['Zm_mAbs'] * np.exp(1j * df_merged['Zm_mPhi'])
comp2 = df_merged['Zm_mRe'] + (1j * df_merged['Zm_mIm'])
"""



# *-* coding: utf-8 *-*
"""Import data in the CRMod/CRTomo format
"""
import os
from glob import glob

import numpy as np
import pandas as pd

from reda.importers.utils.decorators import enable_result_transforms


def load_mod_file(filename):
    """Load a .mod file (sometimes also called volt.dat or data.crt). This file
    contains the number of measurements in the first line, and in the following
    lines measurements. Each line consists of 4 columns:

    * the first column contains the current injection electrodes a and b,
      stored as one integer using the equation: a * 1e4 + b
    * the second column contains the voltage measurement electrodes m and n,
      stored as one integer using the equation: m * 1e4 + n
    * the third column contains the measured resistance [Ohm]
    * the fourth column contains the measured phase value [mrad]

    Parameters
    ----------
    filename : str
        Path of filename to import

    Returns
    -------
    df : pandas.DataFrame
        A reda-conform DataFrame

    Examples
    --------

        import reda
        import reda.importers.crtomo as cexp
        df = cexp.load_mod_file('volt_01_0.1Hz.crt')
        ert = reda.ERT(data=df)

    """
    df_raw = pd.read_csv(
        filename,
        skiprows=1,
        delim_whitespace=True,
        names=['ab', 'mn', 'r', 'rpha']
    )
    df_raw['Zt'] = df_raw['r'] * np.exp(1j * df_raw['rpha'] / 1000.0)
    # ok, this is tricky: the preceding line, computing Zt, always makes sure
    # that the resulting complex number is located in the upper right quadrant
    # of the complex plane, thereby implicitly correcting for any negative
    # K-factor. We do not want this to ensure data integrity (i.e., electrode
    # numbers in the abmn columns would need to be changed, too). Therefore, if
    # r < 0, switch the sign of Zt.
    r_smaller_0 = df_raw['r'] < 0
    df_raw.loc[r_smaller_0, 'Zt'] *= -1
    # print('crtomo import')
    # import IPython
    # IPython.embed()
    df_raw['a'] = np.floor(df_raw['ab'] / 1e4).astype(int)
    df_raw['b'] = (df_raw['ab'] % 1e4).astype(int)
    df_raw['m'] = np.floor(df_raw['mn'] / 1e4).astype(int)
    df_raw['n'] = (df_raw['mn'] % 1e4).astype(int)

    df = df_raw.drop(['ab', 'mn'], axis=1)
    if np.all(np.isnan(df['rpha'])):
        df = df.drop(['rpha', 'Zt'], axis=1)

    return df


@enable_result_transforms
def load_seit_data(directory, frequency_file='frequencies.dat',
                   data_prefix='volt_', **kwargs):
    """Load sEIT data from data directory. This function loads data previously
    exported from reda using reda.exporters.crtomo.write_files_to_directory

    Parameters
    ----------
    directory : string
        input directory
    frequency_file : string, optional
        file (located in directory) that contains the frequencies
    data_prefix: string, optional
        for each frequency a corresponding data file must be present in the
        input directory. Frequencies and files are matched by sorting the
        frequencies AND the filenames, retrieved using glob and the
        data_prefix

    Returns
    -------
    df : pandas.DataFrame
        A DataFrame suitable for the sEIT container
    electrodes : None
        No electrode data is imported
    topography : None
        No topography data is imported

    """
    frequencies = np.loadtxt(directory + os.sep + frequency_file)
    data_files = sorted(glob(directory + os.sep + data_prefix + '*'))
    # check that the number of frequencies matches the number of data files
    if frequencies.size != len(data_files):
        raise Exception(
            'number of frequencies does not match number of data files')

    # load data
    data_list = []
    for frequency, filename in zip(frequencies, data_files):
        subdata = load_mod_file(filename)
        subdata['frequency'] = frequency
        data_list.append(subdata)
    df = pd.concat(data_list)
    return df, None, None



# -*- coding: utf-8 -*-
""" Work with result files from the EIT-40 tomograph (also called medusa).

NOTE that the results for EIT40 and EIT160 are, at this time, slightly
different.

Data structure of .mat files:

    EMD(n).fm frequency
    EMD(n).Time point of time of this measurement
    EMD(n).ni number of the two excitation electrodes (not the channel number)
    EMD(n).nu number of the two potential electrodes (not the channel number)
    EMD(n).Zt3 array with the transfer impedances (repetition measurement)
    EMD(n).nni number of injection
    EMD(n).cni number of channels used to inject current
    EMD(n).cnu number of channels used to measure voltage
    EMD(n).Is3 injected current (A) (calibrated)
    EMD(n).II3 leakage current (A)
    EMD(n).Yg1 Admitance of first injection path
    EMD(n).Yg2 Admitance of second injection path
    EMD(n).As3 Voltages at shunt resistors (defined in .mcf files: NA1 - NA2)
    EMD(n).Zg3 Impedance between injection electrodes


Import pipeline:

- read single-potentials from .mat file
- read quadrupoles from separate file or provide numpy array
- compute mean of three impedance measurement repetitions (Z1-Z3) for each ABM
- compute quadrupole impedance via superposition using

    - a) averaged Z-values
    - b) the single repetitions Z1-Z3

* (I think we don't need the next step because of np.arctan2)
  check for correct quadrant in phase values, correct if necessary (is this
  required if we use the arctan2 function?)

* compute variance/standard deviation from the repetition values

* should we provide a time delta between the two measurements?

"""
import numpy as np
import scipy.io as sio
import pandas as pd
import datetime

import reda.utils.geometric_factors as redaK


def _add_rhoa(df, spacing):
    """a simple wrapper to compute K factors and add rhoa
    """
    df['k'] = redaK.compute_K_analytical(df, spacing=spacing)
    df['rho_a'] = df['r'] * df['k']
    if 'Zt' in df.columns:
        df['rho_a_complex'] = df['Zt'] * df['k']
    return df


def import_medusa_data(mat_filename, configs):
    """

    """
    df_emd = _read_mat_mnu0(mat_filename)

    # 'configs' can be a numpy array or a filename
    if not isinstance(configs, np.ndarray):
        configs = np.loadtxt(configs).astype(int)

    # construct four-point measurements via superposition
    quadpole_list = []
    index = 0
    for Ar, Br, M, N in configs:
        # check if this config is suitable
        if np.unique((Ar, Br, M, N)).size != 4:
            print('ignoring', Ar, Br, M, N)
            continue

        print('constructing', Ar, Br, M, N)
        # the order of A and B doesn't concern us
        A = np.min((Ar, Br))
        B = np.max((Ar, Br))

        # first choice: correct ordering
        query_M = df_emd.query('a=={0} and b=={1} and p=={2}'.format(
            A, B, M
        ))
        query_N = df_emd.query('a=={0} and b=={1} and p=={2}'.format(
            A, B, N
        ))

        if query_M.size == 0 or query_N.size == 0:
            print(
                'Could not find suitable injections',
                query_M.size, query_N.size
            )
            continue

        index += 1

        # keep these columns as they are (no subtracting)
        keep_cols = [
            'datetime',
            'frequency',
            'a', 'b',
            'Zg1', 'Zg2', 'Zg3',
            'Is',
            'Il',
            'Zg',
            'Iab',
        ]

        df4 = pd.DataFrame()
        diff_cols = ['Zt', ]
        df4[keep_cols] = query_M[keep_cols]
        for col in diff_cols:
            df4[col] = query_M[col].values - query_N[col].values
        df4['m'] = query_M['p'].values
        df4['n'] = query_N['p'].values

        quadpole_list.append(df4)
    dfn = pd.concat(quadpole_list)

    Rsign = np.sign(dfn['Zt'].real)
    dfn['r'] = Rsign * np.abs(dfn['Zt'])
    dfn['Vmn'] = dfn['r'] * dfn['Iab']
    dfn['rpha'] = np.arctan2(
        np.imag(dfn['Zt'].values),
        np.real(dfn['Zt'].values)
    ) * 1e3

    df_final = dfn.reset_index()
    return df_final


def _read_mat_mnu0(filename):
    """Import a .mat file with single potentials (A B M) into a pandas
    DataFrame

    Also export some variables of the md struct into a separate structure
    """
    print('read_mag_single_file')

    mat = sio.loadmat(filename)

    df_emd = _extract_emd(mat)

    return df_emd


def _average_swapped_current_injections(df):
    AB = df[['a', 'b']].values

    # get unique injections
    abu = np.unique(
        AB.flatten().view(AB.dtype.descr * 2)
    ).view(AB.dtype).reshape(-1, 2)
    # find swapped pairs
    pairs = []
    alone = []
    abul = [x.tolist() for x in abu]
    for ab in abul:
        swap = list(reversed(ab))
        if swap in abul:
            pair = (ab, swap)
            pair_r = (swap, ab)
            if pair not in pairs and pair_r not in pairs:
                pairs.append(pair)
        else:
            alone.append(ab)

    # check that all pairs got assigned
    if len(pairs) * 2 + len(alone) != len(abul):
        print('len(pairs) * 2 == {0}'.format(len(pairs) * 2))
        print(len(abul))
        raise Exception(
            'total numbers of unswapped-swapped matching do not match!'
        )
    if len(pairs) > 0 and len(alone) > 0:
        print(
            'WARNING: Found both swapped configurations and non-swapped ones!'
        )

    delete_slices = []

    # these are the columns that we workon (and that are retained)
    columns = [
        'frequency', 'a', 'b', 'p',
        'Z1', 'Z2', 'Z3',
        'Il1', 'Il2', 'Il3',
        'Is1', 'Is2', 'Is3',
        'Zg1', 'Zg2', 'Zg3',
        'datetime',
    ]
    dtypes = {col: df.dtypes[col] for col in columns}

    X = df[columns].values

    for pair in pairs:
        index_a = np.where(
            (X[:, 1] == pair[0][0]) & (X[:, 2] == pair[0][1])
        )[0]
        index_b = np.where(
            (X[:, 1] == pair[1][0]) & (X[:, 2] == pair[1][1])
        )[0]
        # normal injection
        A = X[index_a, :]
        # swapped injection
        B = X[index_b, :]
        # make sure we have the same ordering in P, frequency
        diff = A[:, [0, 3]] - B[:, [0, 3]]
        if not np.all(diff) == 0:
            raise Exception('Wrong ordering')
        # compute the averages in A
        # the minus stems from the swapped current electrodes
        X[index_a, 4:10] = (A[:, 4:10] - B[:, 4:10]) / 2.0
        X[index_a, 10:16] = (A[:, 10:16] + B[:, 10:16]) / 2.0

        delete_slices.append(
            index_b
        )
    X_clean = np.delete(X, np.vstack(delete_slices), axis=0)
    df_clean = pd.DataFrame(X_clean, columns=columns)
    # for col in columns:
    #   # df_clean[col] = df_clean[col].astype(dtypes[col])
    df_clean = df_clean.astype(dtype=dtypes)
    return df_clean


def _extract_emd(mat):
    emd = mat['EMD'].squeeze()
    # Labview epoch
    epoch = datetime.datetime(1904, 1, 1)

    def convert_epoch(x):
        timestamp = epoch + datetime.timedelta(seconds=x.astype(float))
        return timestamp

    dfl = []
    # loop over frequencies
    for f_id in range(0, emd.size):
        # print('Frequency: ', emd[f_id]['fm'])
        fdata = emd[f_id]
        # fdata_md = md[f_id]

        timestamp = np.atleast_2d(
            [convert_epoch(x) for x in fdata['Time'].squeeze()]
        ).T
        df = pd.DataFrame(
            np.hstack((
                timestamp,
                fdata['ni'],
                fdata['nu'],
                fdata['Z3'],
                fdata['Is3'],
                fdata['Il3'],
                fdata['Zg3'],
            )),
        )
        df.columns = (
            'datetime',
            'a',
            'b',
            'p',
            'Z1',
            'Z2',
            'Z3',
            'Is1',
            'Is2',
            'Is3',
            'Il1',
            'Il2',
            'Il3',
            'Zg1',
            'Zg2',
            'Zg3',
        )

        df['frequency'] = np.ones(df.shape[0]) * fdata['fm'].squeeze()

        # cast to correct type
        df['datetime'] = pd.to_datetime(df['datetime'])
        df['a'] = df['a'].astype(int)
        df['b'] = df['b'].astype(int)
        df['p'] = df['p'].astype(int)

        df['Z1'] = df['Z1'].astype(complex)
        df['Z2'] = df['Z2'].astype(complex)
        df['Z3'] = df['Z3'].astype(complex)

        df['Zg1'] = df['Zg1'].astype(complex)
        df['Zg2'] = df['Zg2'].astype(complex)
        df['Zg3'] = df['Zg3'].astype(complex)

        df['Is1'] = df['Is1'].astype(complex)
        df['Is2'] = df['Is2'].astype(complex)
        df['Is3'] = df['Is3'].astype(complex)

        df['Il1'] = df['Il1'].astype(complex)
        df['Il2'] = df['Il2'].astype(complex)
        df['Il3'] = df['Il3'].astype(complex)

        dfl.append(df)

    df = pd.concat(dfl)

    # average swapped current injections here!
    df = _average_swapped_current_injections(df)

    # sort current injections
    condition = df['a'] > df['b']
    df.loc[condition, ['a', 'b']] = df.loc[condition, ['b', 'a']].values
    # change sign because we changed A and B
    df.loc[condition, ['Z1', 'Z2', 'Z3']] *= -1

    # average of Z1-Z3
    df['Zt'] = np.mean(df[['Z1', 'Z2', 'Z3']].values, axis=1)
    # we need to keep the sign of the real part
    sign_re = df['Zt'].real / np.abs(df['Zt'].real)
    df['r'] = np.abs(df['Zt']) * sign_re
    # df['Zt_std'] = np.std(df[['Z1', 'Z2', 'Z3']].values, axis=1)

    df['Is'] = np.mean(df[['Is1', 'Is2', 'Is3']].values, axis=1)
    df['Il'] = np.mean(df[['Il1', 'Il2', 'Il3']].values, axis=1)
    df['Zg'] = np.mean(df[['Zg1', 'Zg2', 'Zg3']].values, axis=1)

    # "standard" injected current, in [mA]
    df['Iab'] = np.abs(df['Is']) * 1e3
    df['Iab'] = df['Iab'].astype(float)
    # df['Is_std'] = np.std(df[['Is1', 'Is2', 'Is3']].values, axis=1)

    return df


def apply_correction_factors(df, correction_file):
    """Apply correction factors for a pseudo-2D measurement setup. See Weigand
    and Kemna, 2017, Biogeosciences, for detailed information.
    """
    if isinstance(correction_file, (list, tuple)):
        corr_data_raw = np.vstack(
            [np.loadtxt(x) for x in correction_file]
        )
    else:
        corr_data_raw = np.loadtxt(correction_file)
    A = (corr_data_raw[:, 0] / 1e4).astype(int)
    B = (corr_data_raw[:, 0] % 1e4).astype(int)
    M = (corr_data_raw[:, 1] / 1e4).astype(int)
    N = (corr_data_raw[:, 1] % 1e4).astype(int)

    corr_data = np.vstack((A, B, M, N, corr_data_raw[:, 2])).T
    corr_data[:, 0:2] = np.sort(corr_data[:, 0:2], axis=1)
    corr_data[:, 2:4] = np.sort(corr_data[:, 2:4], axis=1)

    if 'frequency' not in df.columns:
        raise Exception(
            'No frequency data found. Are you sure this is a seit data set?'
        )

    gf = df.groupby(['a', 'b', 'm', 'n'])
    for key, item in gf.groups.items():
        # print('key', key)
        # print(item)
        item_norm = np.hstack((np.sort(key[0:2]), np.sort(key[2:4])))
        # print(item_norm)
        index = np.where(
            (corr_data[:, 0] == item_norm[0]) &
            (corr_data[:, 1] == item_norm[1]) &
            (corr_data[:, 2] == item_norm[2]) &
            (corr_data[:, 3] == item_norm[3])
        )[0]
        # print(index, corr_data[index])
        if len(index) == 0:
            print(key)
            # import IPython
            # IPython.embed()
            raise Exception(
                'No correction factor found for this configuration'
            )

        factor = corr_data[index, 4]
        # apply correction factor
        for col in ('r', 'Zt', 'Vmn', 'rho_a'):
            if col in df.columns:
                df.iloc[item, df.columns.get_loc(col)] *= factor
    return corr_data



"""Exporter for BERT and pyGimli to unified data format. See:

https://gitlab.com/resistivity-net/bert#the-unified-data-format
"""
# import numpy as np
import logging

import pandas as pd

from reda.utils import has_multiple_timesteps, split_timesteps

logger = logging.getLogger(__name__)


def export_bert(data, electrodes, filename):
    """Export to unified data format used in pyGIMLi & BERT.

    Multiple timesteps are exported to multiple filenames.

    Parameters
    ----------
    data : :py:class:`pandas.DataFrame`
        DataFrame with at least a, b, m, n and r.
    electrodes : :py:class:`pandas.DataFrame`
        DataFrame with electrode positions.
    filename : str
        String of the output filename.
    """
    assert isinstance(electrodes, pd.DataFrame), \
        'This file format requires electrodes! Cannot proceed without them.'

    # Check for multiple timesteps
    if has_multiple_timesteps(data):
        for i, timestep in enumerate(split_timesteps(data)):
            export_bert(
                timestep, electrodes, filename.replace(".", "_%.3d." % i))

        # TODO: Make ABMN consistent
        # index_full = ert.data.groupby(list("abmn")).groups.keys()
        # g = ert.data.groupby('timestep')
        # q = ert.data.pivot_table(
        #   values='r', index=list("abmn"), columns="timestep", dropna=True)
        # ert.data.reset_index(list("abmn"))

    f = open(filename, 'w')

    f.write("%d\n" % len(electrodes))
    f.write("# ")

    # Make temporary copies for renaming
    electrodes = electrodes.copy()
    data = data.copy()

    electrodes.columns = electrodes.columns.str.lower()
    data.columns = data.columns.str.lower()

    # Remove unnecessary columns and rename according to bert conventions
    # https://gitlab.com/resistivity-net/bert#the-unified-data-format
    cols_to_export = ["a", "b", "m", "n", "u", "i", "r", "rho_a", "error"]
    data.drop(data.columns.difference(cols_to_export), 1, inplace=True)
    data.rename(columns={"rho_a": "rhoa", "error": "err"}, inplace=True)

    for key in electrodes.keys():
        f.write("%s " % key)
    f.write("\n")
    for row in electrodes.itertuples(index=False):
        for val in row:
            f.write("%5.3f " % val)
        f.write("\n")
    f.write("%d\n" % len(data))
    f.write("# ")

    # Make sure that a, b, m, n are the first 4 columns
    columns = data.columns.tolist()
    for c in "abmn":
        columns.remove(c)
    columns = list("abmn") + columns
    data = data[columns]

    for key in data.keys():
        f.write("%s " % key)
    f.write("\n")
    for row in data.itertuples(index=False):
        for i, val in enumerate(row):
            if i < 4:
                f.write("%d " % val)
            else:
                f.write("%E " % val)

        f.write("\n")
    f.close()
